#!/usr/bin/env python3

import abc
import argparse
import distutils.util
import enum
import functools
import glob
import itertools
import logging
import math
import multiprocessing
import os
import pathlib
import pyudev
import re
import shutil
import subprocess
import sys
import urllib.request
import yaml
import platform
import shlex
import psutil
import mmap
import datetime

dry_run_mode = False
def perftune_print(log_msg, *args, **kwargs):
    if dry_run_mode:
        log_msg = "# " + log_msg
    print(log_msg, *args, **kwargs)

def __run_one_command(prog_args, stderr=None, check=True):
    proc = subprocess.Popen(prog_args, stdout = subprocess.PIPE, stderr = stderr)
    outs, errs = proc.communicate()
    outs = str(outs, 'utf-8')

    if check and proc.returncode != 0:
        raise subprocess.CalledProcessError(returncode=proc.returncode, cmd=" ".join(prog_args), output=outs, stderr=errs)

    return outs

def run_one_command(prog_args, stderr=None, check=True):
    if dry_run_mode:
        print(" ".join([shlex.quote(x) for x in prog_args]))
    else:
        __run_one_command(prog_args, stderr=stderr, check=check)

def run_read_only_command(prog_args, stderr=None, check=True):
    return __run_one_command(prog_args, stderr=stderr, check=check)

def run_hwloc_distrib(prog_args):
    """
    Returns a list of strings - each representing a single line of hwloc-distrib output.
    """
    return run_read_only_command(['hwloc-distrib'] + prog_args).splitlines()

def run_hwloc_calc(prog_args):
    """
    Returns a single string with the result of the execution.
    """
    return run_read_only_command(['hwloc-calc'] + prog_args).rstrip()

def run_ethtool(prog_args):
    """
    Returns a list of strings - each representing a single line of ethtool output.
    """
    return run_read_only_command(['ethtool'] + prog_args).splitlines()

def fwriteln(fname, line, log_message, log_errors=True):
    try:
        if dry_run_mode:
            print("echo {} > {}".format(line, fname))
            return
        else:
            with open(fname, 'w') as f:
                f.write(line)
            print(log_message)
    except:
        if log_errors:
            print("{}: failed to write into {}: {}".format(log_message, fname, sys.exc_info()))

def readlines(fname):
    try:
        with open(fname, 'r') as f:
            return f.readlines()
    except:
        print("Failed to read {}: {}".format(fname, sys.exc_info()))
        return []

def fwriteln_and_log(fname, line, log_errors=True):
    msg = "Writing '{}' to {}".format(line, fname)
    fwriteln(fname, line, log_message=msg, log_errors=log_errors)

double_commas_pattern = re.compile(',,')

def set_one_mask(conf_file, mask, log_errors=True):
    if not os.path.exists(conf_file):
        raise Exception("Configure file to set mask doesn't exist: {}".format(conf_file))
    mask = re.sub('0x', '', mask)

    while double_commas_pattern.search(mask):
        mask = double_commas_pattern.sub(',0,', mask)

    msg = "Setting mask {} in {}".format(mask, conf_file)
    fwriteln(conf_file, mask, log_message=msg, log_errors=log_errors)

def distribute_irqs(irqs, cpu_mask, log_errors=True):
    # If IRQs' list is empty - do nothing
    if not irqs:
        return

    for i, mask in enumerate(run_hwloc_distrib(["{}".format(len(irqs)), '--single', '--restrict', cpu_mask])):
        set_one_mask("/proc/irq/{}/smp_affinity".format(irqs[i]), mask, log_errors=log_errors)

def is_process_running(name):
    return len(list(filter(lambda ps_line : not re.search('<defunct>', ps_line), run_read_only_command(['ps', '--no-headers', '-C', name], check=False).splitlines()))) > 0

def restart_irqbalance(banned_irqs):
    """
    Restart irqbalance if it's running and ban it from moving the IRQs from the
    given list.
    """
    config_file = '/etc/default/irqbalance'
    options_key = 'OPTIONS'
    systemd = False
    banned_irqs_list = list(banned_irqs)

    # If there is nothing to ban - quit
    if not banned_irqs_list:
        return

    # return early if irqbalance is not running
    if not is_process_running('irqbalance'):
        perftune_print("irqbalance is not running")
        return

    # If this file exists - this a "new (systemd) style" irqbalance packaging.
    # This type of packaging uses IRQBALANCE_ARGS as an option key name, "old (init.d) style"
    # packaging uses an OPTION key.
    if os.path.exists('/lib/systemd/system/irqbalance.service') or \
        os.path.exists('/usr/lib/systemd/system/irqbalance.service'):
        options_key = 'IRQBALANCE_ARGS'
        systemd = True

    if not os.path.exists(config_file):
        if os.path.exists('/etc/sysconfig/irqbalance'):
            config_file = '/etc/sysconfig/irqbalance'
        elif os.path.exists('/etc/conf.d/irqbalance'):
            config_file = '/etc/conf.d/irqbalance'
            options_key = 'IRQBALANCE_OPTS'
            with open('/proc/1/comm', 'r') as comm:
                systemd = 'systemd' in comm.read()
        else:
            perftune_print("Unknown system configuration - not restarting irqbalance!")
            perftune_print("You have to prevent it from moving IRQs {} manually!".format(banned_irqs_list))
            return

    orig_file = "{}.scylla.orig".format(config_file)

    # Save the original file
    if not dry_run_mode:
        if not os.path.exists(orig_file):
            print("Saving the original irqbalance configuration is in {}".format(orig_file))
            shutil.copyfile(config_file, orig_file)
        else:
            print("File {} already exists - not overwriting.".format(orig_file))

    # Read the config file lines
    cfile_lines = open(config_file, 'r').readlines()

    # Build the new config_file contents with the new options configuration
    perftune_print("Restarting irqbalance: going to ban the following IRQ numbers: {} ...".format(", ".join(banned_irqs_list)))

    # Search for the original options line
    opt_lines = list(filter(lambda line : re.search(r"^\s*{}".format(options_key), line), cfile_lines))
    if not opt_lines:
        new_options = "{}=\"".format(options_key)
    elif len(opt_lines) == 1:
        # cut the last "
        new_options = re.sub(r'"\s*$', "", opt_lines[0].rstrip())
        opt_lines = opt_lines[0].strip()
    else:
        raise Exception("Invalid format in {}: more than one lines with {} key".format(config_file, options_key))

    for irq in banned_irqs_list:
        # prevent duplicate "ban" entries for the same IRQ
        opt = f"--banirq={irq}"
        patt_str = rf"{opt}\Z|{opt}\s"
        if not re.search(patt_str, new_options):
            new_options += f" {opt}"

    new_options += "\""

    if dry_run_mode:
        if opt_lines:
            print("sed -i 's/^{}/#{}/g' {}".format(options_key, options_key, config_file))
        print("echo {} | tee -a {}".format(new_options, config_file))
    else:
        with open(config_file, 'w') as cfile:
            for line in cfile_lines:
                if not re.search(r"^\s*{}".format(options_key), line):
                    cfile.write(line)

            cfile.write(new_options + "\n")

    if systemd:
        perftune_print("Restarting irqbalance via systemctl...")
        run_one_command(['systemctl', 'try-restart', 'irqbalance'])
    else:
        perftune_print("Restarting irqbalance directly (init.d)...")
        run_one_command(['/etc/init.d/irqbalance', 'restart'])

def learn_irqs_from_proc_interrupts(pattern, irq2procline):
    return [ irq for irq, proc_line in filter(lambda irq_proc_line_pair : re.search(pattern, irq_proc_line_pair[1]), irq2procline.items()) ]

def learn_all_irqs_one(irq_conf_dir, irq2procline, xen_dev_name):
    """
    Returns a list of IRQs of a single device.

    irq_conf_dir: a /sys/... directory with the IRQ information for the given device
    irq2procline: a map of IRQs to the corresponding lines in the /proc/interrupts
    xen_dev_name: a device name pattern as it appears in the /proc/interrupts on Xen systems
    """
    msi_irqs_dir_name = os.path.join(irq_conf_dir, 'msi_irqs')
    # Device uses MSI IRQs
    if os.path.exists(msi_irqs_dir_name):
        return os.listdir(msi_irqs_dir_name)

    irq_file_name = os.path.join(irq_conf_dir, 'irq')
    # Device uses INT#x
    if os.path.exists(irq_file_name):
        return [ line.lstrip().rstrip() for line in open(irq_file_name, 'r').readlines() ]

    # No irq file detected
    modalias = open(os.path.join(irq_conf_dir, 'modalias'), 'r').readline()

    # virtio case
    if re.search("^virtio", modalias):
        return list(itertools.chain.from_iterable(
            map(lambda dirname : learn_irqs_from_proc_interrupts(dirname, irq2procline),
                filter(lambda dirname : re.search('virtio', dirname),
                       itertools.chain.from_iterable([ dirnames for dirpath, dirnames, filenames in os.walk(os.path.join(irq_conf_dir, 'driver')) ])))))

    # xen case
    if re.search("^xen:", modalias):
        return learn_irqs_from_proc_interrupts(xen_dev_name, irq2procline)

    return []

def get_irqs2procline_map():
    return { line.split(':')[0].lstrip().rstrip() : line for line in open('/proc/interrupts', 'r').readlines() }


class AutodetectError(Exception):
    pass


def auto_detect_irq_mask(cpu_mask, cores_per_irq_core):
    """
    The logic of auto-detection of what was once a 'mode' is generic and is all about the amount of CPUs and NUMA
    nodes that are present and a restricting 'cpu_mask'.
    This function implements this logic:

    * up to 4 CPU threads: use 'cpu_mask'
    * up to 4 CPU cores (on x86 this would translate to 8 CPU threads): use a single CPU thread out of allowed
    * up to 16 CPU cores: use a single CPU core out of allowed
    * more than 16 CPU cores: use a single CPU core for each 16 CPU cores and distribute them evenly among all
      present NUMA nodes. We will also make sure all NUMA nodes have exactly the same amount of IRQ cores.

    An AutodetectError exception is raised if 'cpu_mask' is defined in a way that there is a different number of threads
    and/or cores among different NUMA nodes. In such a case a user needs to provide
    an IRQ CPUs definition explicitly using 'irq_cpu_mask' parameter.

    :param cpu_mask: CPU mask that defines which out of present CPUs can be considered for tuning
    :param cores_per_irq_core number of cores to allocate a single IRQ core out of, e.g. 6 means allocate a single IRQ
                              core out of every 6 CPU cores.
    :return: CPU mask to bind IRQs to, a.k.a. irq_cpu_mask
    """
    cores_key = 'cores'
    PUs_key = 'PUs'

    # List of NUMA IDs that own CPUs from the given CPU mask
    numa_ids_list = run_hwloc_calc(['-I', 'numa', cpu_mask]).split(",")

    # Let's calculate number of HTs and cores on each NUMA node belonging to the given CPU set
    cores_PUs_per_numa = {} # { <numa_id> : {'cores':  <number of cores>, 'PUs': <number of PUs>}}
    for n in numa_ids_list:
        num_cores = int(run_hwloc_calc(['--restrict', cpu_mask, '--number-of', 'core', f'numa:{n}']))
        num_PUs = int(run_hwloc_calc(['--restrict', cpu_mask, '--number-of', 'PU', f'numa:{n}']))
        cores_PUs_per_numa[n] = {cores_key: num_cores, PUs_key: num_PUs}

    # Let's check if configuration on each NUMA is the same. If it's not then we can't auto-detect the IRQs CPU set
    # and a user needs to provide it explicitly
    num_cores0 = cores_PUs_per_numa[numa_ids_list[0]][cores_key]
    num_PUs0 = cores_PUs_per_numa[numa_ids_list[0]][PUs_key]
    for n in numa_ids_list:
        if cores_PUs_per_numa[n][cores_key] != num_cores0 or cores_PUs_per_numa[n][PUs_key] != num_PUs0:
            raise AutodetectError(f"NUMA{n} has a different configuration from NUMA0 for a given CPU mask {cpu_mask}: "
                                  f"{cores_PUs_per_numa[n][cores_key]}:{cores_PUs_per_numa[n][PUs_key]} vs "
                                  f"{num_cores0}:{num_PUs0}. Auto-detection of IRQ CPUs in not possible. "
                                  f"Please, provide irq_cpu_mask explicitly.")

    # Auto-detection of IRQ CPU set is possible - let's get to it!
    #
    # Total counts for the whole machine
    num_cores = int(run_hwloc_calc(['--restrict', cpu_mask, '--number-of', 'core', 'machine:0']))
    num_PUs = int(run_hwloc_calc(['--restrict', cpu_mask, '--number-of', 'PU', 'machine:0']))

    if num_PUs <= 4:
        return cpu_mask
    elif num_cores <= 4:
        return run_hwloc_calc(['--restrict', cpu_mask, 'PU:0'])
    elif num_cores <= cores_per_irq_core:
        return run_hwloc_calc(['--restrict', cpu_mask, 'core:0'])
    else:
        # Big machine.
        # Let's allocate a full core out of every cores_per_irq_core cores.
        # Let's distribute IRQ cores among present NUMA nodes
        num_irq_cores = len(numa_ids_list) * math.ceil(num_cores0 / cores_per_irq_core)
        hwloc_args = []
        numa_cores_count = {n: 0 for n in numa_ids_list}
        added_cores = 0
        while added_cores < num_irq_cores:
            for numa in numa_ids_list:
                hwloc_args.append(f"node:{numa}.core:{numa_cores_count[numa]}")
                added_cores += 1
                numa_cores_count[numa] += 1

                if added_cores >= num_irq_cores:
                    break

        return run_hwloc_calc(['--restrict', cpu_mask] + hwloc_args)


################################################################################
class PerfTunerBase(metaclass=abc.ABCMeta):
    def __init__(self, args):
        self.__args = args
        self.__args.cpu_mask = run_hwloc_calc(['--restrict', self.__args.cpu_mask, 'all'])
        self.__mode = None
        self.__compute_cpu_mask = None

        if self.args.mode:
            self.mode = PerfTunerBase.SupportedModes[self.args.mode]
        elif args.irq_cpu_mask:
            self.irqs_cpu_mask = args.irq_cpu_mask
        else:
            self.irqs_cpu_mask = auto_detect_irq_mask(self.cpu_mask, self.cores_per_irq_core)

        self.__is_aws_i3_nonmetal_instance = None
        self.__metadata_token_value = None
        self.__metadata_token_time = None

#### Public methods ##########################
    class CPUMaskIsZeroException(Exception):
        """Thrown if CPU mask turns out to be zero"""
        pass

    class SupportedModes(enum.IntEnum):
        """
        Modes are ordered from the one that cuts the biggest number of CPUs
        from the compute CPUs' set to the one that takes the smallest ('mq' doesn't
        cut any CPU from the compute set).

        This fact is used when we calculate the 'common quotient' mode out of a
        given set of modes (e.g. default modes of different Tuners) - this would
        be the smallest among the given modes.
        """
        sq_split = 0
        sq = 1
        mq = 2

        # Note: no_irq_restrictions should always have the greatest value in the enum since it's the least restricting mode.
        no_irq_restrictions = 9999

        @staticmethod
        def names():
            return PerfTunerBase.SupportedModes.__members__.keys()

        @staticmethod
        def combine(modes):
            """
            :param modes: a set of modes of the PerfTunerBase.SupportedModes type
            :return: the mode that is the "common ground" for a given set of modes.
            """

            # Perform an explicit cast in order to verify that the values in the 'modes' are compatible with the
            # expected PerfTunerBase.SupportedModes type.
            return min([PerfTunerBase.SupportedModes(m) for m in modes])

    @staticmethod
    def cpu_mask_is_zero(cpu_mask):
        """
        The cpu_mask is a comma-separated list of 32-bit hex values with possibly omitted zero components,
        e.g. 0xffff,,0xffff
        We want to estimate if the whole mask is all-zeros.
        :param cpu_mask: hwloc-calc generated CPU mask
        :return: True if mask is zero, False otherwise
        """
        for cur_cpu_mask in cpu_mask.split(','):
            if cur_cpu_mask and int(cur_cpu_mask, 16) != 0:
                return False

        return True

    @staticmethod
    def compute_cpu_mask_for_mode(mq_mode, cpu_mask):
        mq_mode = PerfTunerBase.SupportedModes(mq_mode)

        if mq_mode == PerfTunerBase.SupportedModes.sq:
            # all but CPU0
            compute_cpu_mask = run_hwloc_calc([cpu_mask, '~PU:0'])
        elif mq_mode == PerfTunerBase.SupportedModes.sq_split:
            # all but CPU0 and its HT siblings
            compute_cpu_mask = run_hwloc_calc([cpu_mask, '~core:0'])
        elif mq_mode == PerfTunerBase.SupportedModes.mq:
            # all available cores
            compute_cpu_mask = cpu_mask
        elif mq_mode == PerfTunerBase.SupportedModes.no_irq_restrictions:
            # all available cores
            compute_cpu_mask = cpu_mask
        else:
            raise Exception("Unsupported mode: {}".format(mq_mode))

        if PerfTunerBase.cpu_mask_is_zero(compute_cpu_mask):
            raise PerfTunerBase.CPUMaskIsZeroException("Bad configuration mode ({}) and cpu-mask value ({}): this results in a zero-mask for compute".format(mq_mode.name, cpu_mask))

        return compute_cpu_mask

    @staticmethod
    def irqs_cpu_mask_for_mode(mq_mode, cpu_mask):
        mq_mode = PerfTunerBase.SupportedModes(mq_mode)
        irqs_cpu_mask = 0

        if mq_mode != PerfTunerBase.SupportedModes.mq and mq_mode != PerfTunerBase.SupportedModes.no_irq_restrictions:
            irqs_cpu_mask = run_hwloc_calc([cpu_mask, "~{}".format(PerfTunerBase.compute_cpu_mask_for_mode(mq_mode, cpu_mask))])
        else: # mq_mode == PerfTunerBase.SupportedModes.mq or mq_mode == PerfTunerBase.SupportedModes.no_irq_restrictions
            # distribute equally between all available cores
            irqs_cpu_mask = cpu_mask

        if PerfTunerBase.cpu_mask_is_zero(irqs_cpu_mask):
            raise PerfTunerBase.CPUMaskIsZeroException("Bad configuration mode ({}) and cpu-mask value ({}): this results in a zero-mask for IRQs".format(mq_mode.name, cpu_mask))

        return irqs_cpu_mask

    @property
    def mode(self):
        """
        Return the configuration mode
        """
        return self.__mode

    @mode.setter
    def mode(self, new_mode):
        """
        Set the new configuration mode and recalculate the corresponding masks.
        """
        # Make sure the new_mode is of PerfTunerBase.AllowedModes type
        self.__mode = PerfTunerBase.SupportedModes(new_mode)
        self.__compute_cpu_mask = PerfTunerBase.compute_cpu_mask_for_mode(self.__mode, self.__args.cpu_mask)
        self.__irq_cpu_mask = PerfTunerBase.irqs_cpu_mask_for_mode(self.__mode, self.__args.cpu_mask)

    @property
    def cpu_mask(self):
        """
        Return the CPU mask we operate on (the total CPU set)
        """

        return self.__args.cpu_mask

    @property
    def cores_per_irq_core(self):
        """
        Return the number of cores we are going to allocate a single IRQ core out of when auto-detecting
        """
        return self.__args.cores_per_irq_core

    @staticmethod
    def min_cores_per_irq_core():
        """
        A minimum value of cores_per_irq_core.
        We don't allocate a full IRQ core if total number of CPU cores is less or equal to 4.
        """
        return 5

    @property
    def compute_cpu_mask(self):
        """
        Return the CPU mask to use for seastar application binding.
        """
        return self.__compute_cpu_mask

    @property
    def irqs_cpu_mask(self):
        """
        Return the mask of CPUs used for IRQs distribution.
        """
        return self.__irq_cpu_mask

    @irqs_cpu_mask.setter
    def irqs_cpu_mask(self, new_irq_cpu_mask):
        self.__irq_cpu_mask = new_irq_cpu_mask

        # Sanity check
        if PerfTunerBase.cpu_mask_is_zero(self.__irq_cpu_mask):
            raise PerfTunerBase.CPUMaskIsZeroException("Bad configuration: zero IRQ CPU mask is given")

        if run_hwloc_calc([self.__irq_cpu_mask]) == run_hwloc_calc([self.cpu_mask]):
            # Special case: if IRQ CPU mask is the same as total CPU mask - set a Compute CPU mask to cpu_mask
            self.__compute_cpu_mask = self.cpu_mask
        else:
            # Otherwise, a Compute CPU mask is a CPU mask without IRQ CPU mask bits
            self.__compute_cpu_mask = run_hwloc_calc([self.cpu_mask, f"~{self.__irq_cpu_mask}"])

        # Sanity check
        if PerfTunerBase.cpu_mask_is_zero(self.__compute_cpu_mask):
            raise PerfTunerBase.CPUMaskIsZeroException(
                f"Bad configuration: cpu_maks:{self.cpu_mask}, irq_cpu_mask:{self.__irq_cpu_mask}: "
                f"results in a zero-mask for compute")

    @property
    def is_aws_i3_non_metal_instance(self):
        """
        :return: True if we are running on the AWS i3.nonmetal instance, e.g. i3.4xlarge
        """
        if self.__is_aws_i3_nonmetal_instance is None:
            self.__check_host_type()

        return self.__is_aws_i3_nonmetal_instance

    @property
    def args(self):
        return self.__args

    @property
    def irqs(self):
        return self._get_irqs()

#### "Protected"/Public (pure virtual) methods ###########
    @abc.abstractmethod
    def tune(self):
        pass


    @abc.abstractmethod
    def _get_irqs(self):
        """
        Return the iteratable value with all IRQs to be configured.
        """
        pass

#### Private methods ############################
    @property
    def __ec2_metadata_base_url(self):
        return "http://169.254.169.254/latest/"

    @property
    def __metadata_token(self):
        """
        Refresh IMDSv2 session token if it necessary, and return current token
        :return: current session token
        """
        token_ttl = 21600
        update_token = False
        if not self.__metadata_token_value:
            update_token = True
        else:
            time_diff = datetime.datetime.now() - self.__metadata_token_time
            time_diff_sec = int(time_diff.total_seconds())
            if time_diff_sec >= token_ttl - 120:
                update_token = True
        if update_token:
            self.__metadata_token_time = datetime.datetime.now()
            req = urllib.request.Request(self.__ec2_metadata_base_url + "api/token", headers={"X-aws-ec2-metadata-token-ttl-seconds": token_ttl}, method="PUT")
            with urllib.request.urlopen(req, timeout=0.1) as res:
                self.__metadata_token_value = res.read().decode()
        return self.__metadata_token_value

    def __get_instance_metadata(self, path):
        """
        Get a parameter from EC2 Metadata server
        :param path: metadata path to access for
        :return: received metadata as a string
        """
        req = urllib.request.Request(self.__ec2_metadata_base_url + 'meta-data/' + path, headers={"X-aws-ec2-metadata-token": self.__metadata_token})
        with urllib.request.urlopen(req, timeout=0.1) as res:
            return res.read().decode()

    def __check_host_type(self):
        """
        Check if we are running on the AWS i3 nonmetal instance.
        If yes, set self.__is_aws_i3_nonmetal_instance to True, and to False otherwise.
        """
        try:
            aws_instance_type = self.__get_instance_metadata('instance-type')
            if re.match(r'^i3\.((?!metal)\w)+$', aws_instance_type):
                self.__is_aws_i3_nonmetal_instance = True
            else:
                self.__is_aws_i3_nonmetal_instance = False

            return
        except (urllib.error.URLError, ConnectionError, TimeoutError):
            # Non-AWS case
            pass
        except:
            logging.warning("Unexpected exception while attempting to access AWS meta server: {}".format(sys.exc_info()[0]))

        self.__is_aws_i3_nonmetal_instance = False

#################################################
class NetPerfTuner(PerfTunerBase):
    def __init__(self, args):
        super().__init__(args)

        self.nics=args.nics

        self.__nic_is_bond_iface = NetPerfTuner.__get_bond_ifaces()
        self.__nic_is_vlan_iface = NetPerfTuner.__get_vlan_ifaces()
        self.__slaves = self.__learn_slaves()

        # check that self.nics contain a HW device or a supported composite interface
        self.__check_nics()

        # Fetch IRQs related info
        self.__get_irqs_info()


#### Public methods ############################
    def tune(self):
        """
        Tune the networking server configuration.
        """
        for nic in self.nics:
            if self.nic_is_hw_iface(nic):
                perftune_print("Setting a physical interface {}...".format(nic))
                self.__setup_one_hw_iface(nic)
            else:
                perftune_print(f"Setting a {nic} {'bond' if self.nic_is_bond_iface(nic) else 'VLAN'} interface...")
                self.__setup_composite_iface(nic)

        # Increase the socket listen() backlog
        fwriteln_and_log('/proc/sys/net/core/somaxconn', '4096')

        # Increase the maximum number of remembered connection requests, which are still
        # did not receive an acknowledgment from connecting client.
        fwriteln_and_log('/proc/sys/net/ipv4/tcp_max_syn_backlog', '4096')

        self.tune_tcp_mem()

    def tune_tcp_mem(self):
        page_size = mmap.PAGESIZE
        total_mem = psutil.virtual_memory().total
        # We only tune for physical memory since tcp_mem is virtualized
        def to_pages(bytes):
            return math.ceil(bytes / page_size)
        max = total_mem * self.args.tcp_mem_fraction
        fwriteln_and_log('/proc/sys/net/ipv4/tcp_mem', f"{to_pages(max / 2)} {to_pages(max * 2/3)} {to_pages(max)}")

    def nic_is_bond_iface(self, nic):
        return self.__nic_is_bond_iface.get(nic, False)

    def nic_is_vlan_iface(self, nic):
        return self.__nic_is_vlan_iface.get(nic, False)

    def nic_is_composite_iface(self, nic):
        return self.nic_is_bond_iface(nic) or self.nic_is_vlan_iface(nic)

    def nic_exists(self, nic):
        return self.__iface_exists(nic)

    def nic_is_hw_iface(self, nic):
        return self.__dev_is_hw_iface(nic)

    def slaves(self, nic):
        """
        Returns an iterator for all slaves of the nic.
        If agrs.nic is not a composite interface an attempt to use the returned iterator
        will immediately raise a StopIteration exception - use nic_is_composite_iface() check to avoid this.
        """
        return iter(self.__slaves[nic])

#### Protected methods ##########################
    def _get_irqs(self):
        """
        Returns the iterator for all IRQs that are going to be configured (according to args.nics parameter).
        For instance, for a bonding interface that's going to include IRQs of all its slaves.
        """
        return itertools.chain.from_iterable(self.__nic2irqs.values())

#### Private methods ############################
    def __get_irqs_info(self):
        self.__irqs2procline = get_irqs2procline_map()
        self.__nic2irqs = self.__learn_irqs()

    @property
    def __rfs_table_size(self):
        return 32768

    def __check_nics(self):
        """
        Checks that self.nics are supported interfaces
        """
        for nic in self.nics:
            if not self.nic_exists(nic):
                raise Exception("Device {} does not exist".format(nic))
            if not self.nic_is_hw_iface(nic) and not self.nic_is_composite_iface(nic):
                raise Exception("Not supported virtual device {}".format(nic))

    def __get_irqs_one(self, iface):
        """
        Returns the list of IRQ numbers for the given interface.
        """
        return self.__nic2irqs[iface]

    def __setup_rfs(self, iface):
        rps_limits = glob.glob("/sys/class/net/{}/queues/*/rps_flow_cnt".format(iface))
        one_q_limit = int(self.__rfs_table_size / len(rps_limits))

        # If RFS feature is not present - get out
        try:
            run_one_command(['sysctl', 'net.core.rps_sock_flow_entries'])
        except:
            return

        # Enable RFS
        perftune_print("Setting net.core.rps_sock_flow_entries to {}".format(self.__rfs_table_size))
        run_one_command(['sysctl', '-w', 'net.core.rps_sock_flow_entries={}'.format(self.__rfs_table_size)])

        # Set each RPS queue limit
        for rfs_limit_cnt in rps_limits:
            msg = "Setting limit {} in {}".format(one_q_limit, rfs_limit_cnt)
            fwriteln(rfs_limit_cnt, "{}".format(one_q_limit), log_message=msg)

        # Enable/Disable ntuple filtering HW offload on the NIC. This is going to enable/disable aRFS on NICs supporting
        # aRFS since ntuple is pre-requisite for an aRFS feature.
        # If no explicit configuration has been requested enable ntuple (and thereby aRFS) only in MQ mode.
        #
        # aRFS acts similar to (SW) RFS: it places a TCP packet on a HW queue that it supposed to be "close" to an
        # application thread that sent a packet on the same TCP stream.
        #
        # For instance if a given TCP stream was sent from CPU3 then the next Rx packet is going to be placed in an Rx
        # HW queue which IRQ affinity is set to CPU3 or otherwise to the one with affinity close enough to CPU3.
        #
        # Read more here: https://access.redhat.com/documentation/en-us/red_hat_enterprise_linux/6/html/performance_tuning_guide/network-acc-rfs
        #
        # Obviously it would achieve the best result if there is at least one Rx HW queue with an affinity set to each
        # application threads that handle TCP.
        #
        # And, similarly, if we know in advance that there won't be any such HW queue (sq and sq_split modes) - there is
        # no sense enabling aRFS.
        op = "Enable"
        value = 'on'

        if (self.args.enable_arfs is None and self.irqs_cpu_mask != self.cpu_mask) or self.args.enable_arfs is False:
            op = "Disable"
            value = 'off'

        ethtool_msg = "{} ntuple filtering HW offload for {}...".format(op, iface)

        if dry_run_mode:
            perftune_print(ethtool_msg)
            run_one_command(['ethtool','-K', iface, 'ntuple', value], stderr=subprocess.DEVNULL)
        else:
            try:
                print("Trying to {} ntuple filtering HW offload for {}...".format(op.lower(), iface), end='')
                run_one_command(['ethtool','-K', iface, 'ntuple', value], stderr=subprocess.DEVNULL)
                print("ok")
            except:
                print("not supported")

    def __setup_rps(self, iface, mask):
        for one_rps_cpus in self.__get_rps_cpus(iface):
            set_one_mask(one_rps_cpus, mask)

        self.__setup_rfs(iface)

    def __setup_xps(self, iface):
        xps_cpus_list = glob.glob("/sys/class/net/{}/queues/*/xps_cpus".format(iface))
        masks = run_hwloc_distrib(["{}".format(len(xps_cpus_list))])

        for i, mask in enumerate(masks):
            set_one_mask(xps_cpus_list[i], mask)

    def __iface_exists(self, iface):
        if len(iface) == 0:
            return False
        return os.path.exists("/sys/class/net/{}".format(iface))

    def __dev_is_hw_iface(self, iface):
        return os.path.exists("/sys/class/net/{}/device".format(iface))

    @staticmethod
    def __get_bond_ifaces():
        if not os.path.exists('/sys/class/net/bonding_masters'):
            return {}

        bond_dict = {}
        for line in open('/sys/class/net/bonding_masters', 'r').readlines():
            for nic in line.split():
                bond_dict[nic] = True

        return bond_dict

    @staticmethod
    def __get_vlan_ifaces():
        # Each VLAN interface is going to have a corresponding entry in /proc/net/vlan/ directory
        return {pathlib.PurePath(pathlib.Path(f)).name: True
                for f in filter(lambda vlan_name: vlan_name != "/proc/net/vlan/config", glob.glob("/proc/net/vlan/*"))}

    def __learn_slaves_one(self, nic):
        """
        Learn underlying physical devices a given NIC

        :param nic: An interface to search slaves for
        """
        slaves_list = set()
        top_slaves_list = set()

        if self.nic_is_bond_iface(nic):
            top_slaves_list = set(itertools.chain.from_iterable(
                [line.split() for line in open("/sys/class/net/{}/bonding/slaves".format(nic), 'r').readlines()]))
        elif self.nic_is_vlan_iface(nic):
            # VLAN interfaces have a symbolic link 'lower_<parent_interface_name>' under
            # /sys/class/net/<VLAN interface name>.
            #
            # For example:
            #
            # lrwxrwxrwx  1 root root    0 Jul  5 18:38 lower_eno1 -> ../../../pci0000:00/0000:00:1f.6/net/eno1/
            #
            top_slaves_list = set([pathlib.PurePath(pathlib.Path(f).resolve()).name
                                   for f in glob.glob(f"/sys/class/net/{nic}/lower_*")])

        # Slaves can be themselves bond or VLAN devices: let's descend (DFS) all the way down to get physical devices.
        # Bond slaves can't be VLAN interfaces but VLAN interface parent device can be a bond interface.
        # Bond slaves can also be bonds.
        # For simplicity let's not discriminate.
        for s in top_slaves_list:
            if self.nic_is_composite_iface(s):
                slaves_list |= self.__learn_slaves_one(s)
            else:
                slaves_list.add(s)

        return slaves_list

    def __learn_slaves(self):
        """
        Resolve underlying physical devices for interfaces we are requested to configure
        """
        slaves_list_per_nic = {}
        for nic in self.nics:
            current_slaves = self.__learn_slaves_one(nic)
            if current_slaves:
                slaves_list_per_nic[nic] = list(current_slaves)

        return slaves_list_per_nic

    def __intel_irq_to_queue_idx(self, irq):
        """
        Return the HW queue index for a given IRQ for Intel NICs in order to sort the IRQs' list by this index.

        Intel's fast path IRQs have the following name convention:
             <bla-bla>-TxRx-<queue index>

        Intel NICs also have the IRQ for Flow Director (which is not a regular fast path IRQ) whose name looks like
        this:
             <bla-bla>:fdir-TxRx-<index>

        We want to put the Flow Director's IRQ at the end of the sorted list of IRQs.

        :param irq: IRQ number
        :return: HW queue index for Intel NICs and sys.maxsize for all other NICs
        """
        intel_fp_irq_re = re.compile(r"-TxRx-(\d+)")
        fdir_re = re.compile(r"fdir-TxRx-\d+")

        m = intel_fp_irq_re.search(self.__irqs2procline[irq])
        m1 = fdir_re.search(self.__irqs2procline[irq])
        if m and not m1:
            return int(m.group(1))
        else:
            return sys.maxsize

    def __mlx_irq_to_queue_idx(self, irq):
        """
        Return the HW queue index for a given IRQ for Mellanox NICs in order to sort the IRQs' list by this index.

        Mellanox NICs have the IRQ which name looks like
        this:
        mlx5_comp23
             mlx5_comp<index>
        or this:
        mlx4-6
             mlx4-<index>

        :param irq: IRQ number
        :return: HW queue index for Mellanox NICs and sys.maxsize for all other NICs
        """
        mlx5_fp_irq_re = re.compile(r"mlx5_comp(\d+)")
        mlx4_fp_irq_re = re.compile(r"mlx4-(\d+)")

        m5 = mlx5_fp_irq_re.search(self.__irqs2procline[irq])
        if m5:
            return int(m5.group(1))
        else:
            m4 = mlx4_fp_irq_re.search(self.__irqs2procline[irq])
            if m4:
                return int(m4.group(1))

        return sys.maxsize

    def __virtio_irq_to_queue_idx(self, irq):
        """
        Return the HW queue index for a given IRQ for VIRTIO in order to sort the IRQs' list by this index.

        VIRTIO NICs have the IRQ's name that looks like this:
        Queue K of a device virtioY, where Y is some integer is comprised of 2 IRQs
        with following names:
          * Tx IRQ:
               virtioY-output.K
          * Rx IRQ:
               virtioY-input.K

        :param irq: IRQ number
        :return: HW queue index for VIRTIO fast path IRQ and sys.maxsize for all other IRQs
        """
        virtio_fp_re = re.compile(r"virtio\d+-(input|output)\.(\d+)$")

        virtio_fp_irq = virtio_fp_re.search(self.__irqs2procline[irq])
        if virtio_fp_irq:
            return int(virtio_fp_irq.group(2))

        return sys.maxsize


    def __get_driver_name(self, iface):
        """
        :param iface: Interface to check
        :return: driver name from ethtool
        """

        driver_name = ''
        ethtool_i_lines = run_ethtool(['-i', iface])
        driver_re = re.compile("driver:")
        driver_lines = list(filter(lambda one_line: driver_re.search(one_line), ethtool_i_lines))

        if driver_lines:
            if len(driver_lines) > 1:
                raise Exception("More than one 'driver:' entries in the 'ethtool -i {}' output. Unable to continue.".format(iface))

            driver_name = driver_lines[0].split()[1].strip()

        return driver_name

    def __learn_irqs_one(self, iface):
        """
        This is a slow method that is going to read from the system files. Never
        use it outside the initialization code. Use __get_irqs_one() instead.

        Filter the fast path queues IRQs from the __get_all_irqs_one() result according to the known
        patterns.
        Right now we know about the following naming convention of the fast path queues vectors:
          - Intel:    <bla-bla>-TxRx-<bla-bla>
          - Broadcom: <bla-bla>-fp-<bla-bla>
          - ena:      <bla-bla>-Tx-Rx-<bla-bla>
          - Mellanox: for mlx4
                      mlx4-<queue idx>@<bla-bla>
                      or for mlx5
                      mlx5_comp<queue idx>@<bla-bla>
          - VIRTIO: virtioN-[input|output].D

        So, we will try to filter the etries in /proc/interrupts for IRQs we've got from get_all_irqs_one()
        according to the patterns above.

        If as a result all IRQs are filtered out (if there are no IRQs with the names from the patterns above) then
        this means that the given NIC uses a different IRQs naming pattern. In this case we won't filter any IRQ.

        Otherwise, we will use only IRQs which names fit one of the patterns above.

        For NICs with a limited number of Rx queues the IRQs that handle Rx are going to be at the beginning of the
        list.
        """
        # filter 'all_irqs' to only reference valid keys from 'irqs2procline' and avoid an IndexError on the 'irqs' search below
        all_irqs = set(learn_all_irqs_one("/sys/class/net/{}/device".format(iface), self.__irqs2procline, iface)).intersection(self.__irqs2procline.keys())
        fp_irqs_re = re.compile(r"-TxRx-|-fp-|-Tx-Rx-|mlx4-\d+@|mlx5_comp\d+@|virtio\d+-(input|output)")
        irqs = sorted(list(filter(lambda irq : fp_irqs_re.search(self.__irqs2procline[irq]), all_irqs)))
        if irqs:
            irqs.sort(key=self.__get_irq_to_queue_idx_functor(iface))
            return irqs
        else:
            return list(all_irqs)

    def __get_irq_to_queue_idx_functor(self, iface):
        """
        Get a functor returning a queue index for a given IRQ.
        This functor is needed for NICs that are known to not release IRQs when the number of Rx
        channels is reduced or have extra IRQs for non-RSS channels.

        Therefore, for these NICs we need a functor that would allow us to pick IRQs that belong to channels that are
        going to handle TCP traffic: first X channels, where the value of X depends on the NIC's type and configuration.

        For others, e.g. ENA, or Broadcom, which are only going to allocate IRQs that belong to TCP handling channels,
        we don't really need to sort them as long as we filter fast path IRQs and distribute them evenly among IRQ CPUs.

        :param iface: NIC's interface name, e.g. eth19
        :return: A functor that returns a queue index for a given IRQ if a mapping is known
                 or a constant big integer value if mapping is unknown.
        """
        # There are a few known drivers for which we know how to get a queue index from an IRQ name in /proc/interrupts
        driver_name = self.__get_driver_name(iface)

        # Every functor returns a sys.maxsize for an unknown driver IRQs.
        # So, choosing Intel's as a default is as good as any other.
        irq_to_idx_func = self.__intel_irq_to_queue_idx
        if driver_name.startswith("mlx"):
            irq_to_idx_func = self.__mlx_irq_to_queue_idx
        elif driver_name.startswith("virtio"):
            irq_to_idx_func = self.__virtio_irq_to_queue_idx

        return irq_to_idx_func

    def __irq_lower_bound_by_queue(self, iface, irqs, queue_idx):
        """
        Get the index of the first element in irqs array which queue is greater or equal to a given index.
        IRQs array is supposed to be sorted by queues numbers IRQs belong to.

        There are additional assumptions:
          * IRQs array items queue numbers are monotonically not decreasing, and if it increases then it increases by
            one.
          * Queue indexes are numbered starting from zero.

        :param irqs: IRQs array sorted by queues numbers IRQs belong to
        :param queue_idx: Queue index to partition by
        :return: The first index in the IRQs array that corresponds to a queue number greater or equal to a given index
                 which is at least queue_idx. If there is no such IRQ - returns len(irqs).
        """
        irq_to_idx_func = self.__get_irq_to_queue_idx_functor(iface)

        if queue_idx < len(irqs):
            for idx in range(queue_idx, len(irqs)):
                if irq_to_idx_func(irqs[idx]) >= queue_idx:
                    return idx

        return len(irqs)

    def __learn_irqs(self):
        """
        This is a slow method that is going to read from the system files. Never
        use it outside the initialization code.
        """
        nic_irq_dict={}
        for nic in self.nics:
            if self.nic_is_composite_iface(nic):
                for slave in filter(self.__dev_is_hw_iface, self.slaves(nic)):
                    nic_irq_dict[slave] = self.__learn_irqs_one(slave)
            else:
                nic_irq_dict[nic] = self.__learn_irqs_one(nic)
        return nic_irq_dict

    def __get_rps_cpus(self, iface):
        """
        Prints all rps_cpus files names for the given HW interface.

        There is a single rps_cpus file for each RPS queue and there is a single RPS
        queue for each HW Rx queue. Each HW Rx queue should have an IRQ.
        Therefore the number of these files is equal to the number of fast path Rx IRQs for this interface.
        """
        return glob.glob("/sys/class/net/{}/queues/*/rps_cpus".format(iface))

    def __set_rx_channels_count(self, iface, count):
        """
        Try to set the number of Rx channels of a given interface to a given value.

        Rx channels of any NIC can be configured using 'ethtool -L' command using one of the following semantics:

        ethtool -L <iface> rx <count>
        or
        ethtool -L <iface> combined <count>

        If a specific semantics is not supported by a given NIC or if changing the number of channels is not supported
        ethtool is going to return an error.

        Instead of parsing and trying to detect which one of the following semantics a given interface supports we will
        simply try to use both semantics till either one of them succeeds or both fail.


        :param iface: NIC interface name, e.g. eth4
        :param count: number of Rx channels we want to configure
        :return: True if configuration was successful, False otherwise
        """
        options = ["rx", "combined"]
        for o in options:
            try:
                cmd = ['ethtool', '-L', iface, o, f"{count}"]
                perftune_print(f"Executing: {' '.join(cmd)}")
                run_one_command(cmd, stderr=subprocess.DEVNULL)
                return True
            except subprocess.CalledProcessError:
                pass

        return False

    def __setup_one_hw_iface(self, iface):
        # Set Rx channels count to a number of IRQ CPUs unless an explicit count is given
        if self.args.num_rx_queues is not None:
            num_rx_channels = self.args.num_rx_queues
        else:
            num_rx_channels = 0

            # If a mask is wider than 32 bits it's going to be presented as a comma-separated list of 32-bit masks
            # with possibly omitted zero components, e.g. 0x01,0x100,,0x12122
            for m in self.irqs_cpu_mask.split(","):
                if m:
                    num_rx_channels += bin(int(m, 16)).count('1')

        # Let's try setting the number of Rx channels to the number of IRQ CPUs.
        #
        # If we were able to change the number of Rx channels the number of IRQs could have changed.
        # In this case let's refresh IRQs info.
        rx_channels_set = self.__set_rx_channels_count(iface, num_rx_channels)
        if rx_channels_set:
            self.__get_irqs_info()

        max_num_rx_queues = self.__max_rx_queue_count(iface)
        all_irqs = self.__get_irqs_one(iface)

        # Bind the NIC's IRQs according to the configuration mode
        #
        # If this NIC has a limited number of Rx queues then we want to distribute their IRQs separately.
        # For such NICs we've sorted IRQs list so that IRQs that handle Rx are all at the head of the list.
        if rx_channels_set or max_num_rx_queues < len(all_irqs):
            num_rx_queues = self.__get_rx_queue_count(iface)
            # Let's be optimistic during a dry run and assume that the RX channels setting was successful
            if dry_run_mode:
                num_rx_queues = num_rx_channels

            tcp_irqs_lower_bound = self.__irq_lower_bound_by_queue(iface, all_irqs, num_rx_queues)
            perftune_print(f"Distributing IRQs handling Rx and Tx for first {num_rx_queues} channels:")
            distribute_irqs(all_irqs[0:tcp_irqs_lower_bound], self.irqs_cpu_mask)
            perftune_print("Distributing the rest of IRQs")
            distribute_irqs(all_irqs[tcp_irqs_lower_bound:], self.irqs_cpu_mask)
        else:
            perftune_print("Distributing all IRQs")
            distribute_irqs(all_irqs, self.irqs_cpu_mask)

        self.__setup_rps(iface, self.cpu_mask)
        self.__setup_xps(iface)

    def __setup_composite_iface(self, nic):
        """
        Set up the interface which is a bond or a VLAN interface
        :param nic: name of a composite interface to set up
        """
        for slave in self.slaves(nic):
            if self.__dev_is_hw_iface(slave):
                perftune_print("Setting up {}...".format(slave))
                self.__setup_one_hw_iface(slave)
            else:
                perftune_print("Skipping {} (not a physical slave device?)".format(slave))

    def __max_rx_queue_count(self, iface):
        """
        :param iface: Interface to check
        :return: The maximum number of RSS queues for the given interface if there is known limitation and sys.maxsize
        otherwise.

        Networking drivers serving HW with the known maximum RSS queue limitation (due to lack of RSS bits):

        ixgbe:   PF NICs support up to 16 RSS queues.
        ixgbevf: VF NICs support up to 4 RSS queues.
        i40e:    PF NICs support up to 64 RSS queues.
        i40evf:  VF NICs support up to 16 RSS queues.

        """
        driver_to_max_rss = {'ixgbe': 16, 'ixgbevf': 4, 'i40e': 64, 'i40evf': 16}

        driver_name = self.__get_driver_name(iface)
        return driver_to_max_rss.get(driver_name, sys.maxsize)

    def __get_rx_queue_count(self, iface):
        """
        :return: the RSS Rx queues count for the given interface.
        """
        num_irqs = len(self.__get_irqs_one(iface))
        rx_queues_count = len(self.__get_rps_cpus(iface))

        if rx_queues_count == 0:
            rx_queues_count = num_irqs

        return min(self.__max_rx_queue_count(iface), rx_queues_count)



class ClocksourceManager:
    class PreferredClockSourceNotAvailableException(Exception):
        pass

    def __init__(self, args):
        self.__args = args
        self._preferred = {"x86_64": "tsc", "kvm": "kvm-clock"}
        self._arch = self._get_arch()
        self._available_clocksources_file = "/sys/devices/system/clocksource/clocksource0/available_clocksource"
        self._current_clocksource_file = "/sys/devices/system/clocksource/clocksource0/current_clocksource"
        self._recommendation_if_unavailable = { "x86_64": "The tsc clocksource is not available. Consider using a hardware platform where the tsc clocksource is available, or try forcing it withe the tsc=reliable boot option", "kvm": "kvm-clock is not available" }

    def _available_clocksources(self):
        return open(self._available_clocksources_file).readline().split()

    def _current_clocksource(self):
        return open(self._current_clocksource_file).readline().strip()

    def _get_arch(self):
        try:
            virt = run_read_only_command(['systemd-detect-virt']).strip()
            # According to https://www.freedesktop.org/software/systemd/man/latest/systemd-detect-virt.html
            # 'amazon' and 'google' are returned for KVM guests.
            if virt in ["kvm", "amazon", "google"]:
                return "kvm"
        except:
            pass
        return platform.machine()

    def enforce_preferred_clocksource(self):
        fwriteln(self._current_clocksource_file, self._preferred[self._arch], "Setting clocksource to {}".format(self._preferred[self._arch]))

    def preferred(self):
        return self._preferred[self._arch]

    def setting_available(self):
        return self._arch in self._preferred

    def preferred_clocksource_available(self):
        return self._preferred[self._arch] in self._available_clocksources()

    def recommendation_if_unavailable(self):
        return self._recommendation_if_unavailable[self._arch]

class SystemPerfTuner(PerfTunerBase):
    def __init__(self, args):
        super().__init__(args)
        self._clocksource_manager = ClocksourceManager(args)

    def tune(self):
        if self.args.tune_clock:
            if not self._clocksource_manager.setting_available():
                perftune_print("Clocksource setting not available or not needed for this architecture. Not tuning")
            elif not self._clocksource_manager.preferred_clocksource_available():
                perftune_print(self._clocksource_manager.recommendation_if_unavailable())
            else:
                self._clocksource_manager.enforce_preferred_clocksource()

#### Protected methods ##########################
    def _get_irqs(self):
        return []


#################################################
class DiskPerfTuner(PerfTunerBase):
    class SupportedDiskTypes(enum.IntEnum):
        nvme = 0
        non_nvme = 1

    def __init__(self, args):
        super().__init__(args)

        if not (self.args.dirs or self.args.devs):
            raise Exception("'disks' tuning was requested but neither directories nor storage devices were given")

        self.__pyudev_ctx = pyudev.Context()
        self.__dir2disks = self.__learn_directories()
        self.__irqs2procline = get_irqs2procline_map()
        self.__disk2irqs = self.__learn_irqs()
        self.__type2diskinfo = self.__group_disks_info_by_type()

        # sets of devices that have already been tuned
        self.__io_scheduler_tuned_devs = set()
        self.__nomerges_tuned_devs = set()
        self.__write_back_cache_tuned_devs = set()

#### Public methods #############################
    def tune(self):
        """
        Distribute IRQs according to the requested mode (args.mode):
           - Distribute NVMe disks' IRQs equally among all available CPUs.
           - Distribute non-NVMe disks' IRQs equally among designated CPUs or among
             all available CPUs in the 'mq' mode.
        """
        non_nvme_disks, non_nvme_irqs = self.__disks_info_by_type(DiskPerfTuner.SupportedDiskTypes.non_nvme)
        if non_nvme_disks:
            perftune_print("Setting non-NVMe disks: {}...".format(", ".join(non_nvme_disks)))
            distribute_irqs(non_nvme_irqs, self.irqs_cpu_mask)
            self.__tune_disks(non_nvme_disks)
        else:
            perftune_print("No non-NVMe disks to tune")

        nvme_disks, nvme_irqs = self.__disks_info_by_type(DiskPerfTuner.SupportedDiskTypes.nvme)
        if nvme_disks:
            # Linux kernel is going to use IRQD_AFFINITY_MANAGED mode for NVMe IRQs
            # on most systems (currently only AWS i3 non-metal are known to have a
            # different configuration). SMP affinity of an IRQ in this mode may not be
            # changed and an attempt to modify it is going to fail. However right now
            # the only way to determine that IRQD_AFFINITY_MANAGED mode has been used
            # is to attempt to modify IRQ SMP affinity (and fail) therefore we prefer
            # to always do it.
            #
            # What we don't want however is to see annoying errors every time we
            # detect that IRQD_AFFINITY_MANAGED was actually used. Therefore we will only log
            # them in the "verbose" mode or when we run on an i3.nonmetal AWS instance.
            perftune_print("Setting NVMe disks: {}...".format(", ".join(nvme_disks)))
            distribute_irqs(nvme_irqs, self.args.cpu_mask,
                            log_errors=(self.is_aws_i3_non_metal_instance or self.args.verbose))
            self.__tune_disks(nvme_disks)
        else:
            perftune_print("No NVMe disks to tune")

#### Protected methods ##########################
    def _get_irqs(self):
        return itertools.chain.from_iterable(irqs for disks, irqs in self.__type2diskinfo.values())

#### Private methods ############################
    @property
    def __io_schedulers(self):
        """
        :return: An ordered list of IO schedulers that we want to configure. Schedulers are ordered by their priority
        from the highest (left most) to the lowest.
        """
        return ["none", "noop"]

    @property
    def __nomerges(self):
        return '2'

    @property
    def __write_cache_config(self):
        """
        :return: None - if write cache mode configuration is not requested or the corresponding write cache
        configuration value string
        """
        if self.args.set_write_back is None:
            return None

        return "write back" if self.args.set_write_back else "write through"

    def __disks_info_by_type(self, disks_type):
        """
        Returns a tuple ( [<disks>], [<irqs>] ) for the given disks type.
        IRQs numbers in the second list are promised to be unique.
        """
        return self.__type2diskinfo[DiskPerfTuner.SupportedDiskTypes(disks_type)]

    def __nvme_fast_path_irq_filter(self, irq):
        """
        Return True for fast path NVMe IRQs.
        For NVMe device only queues 1-<number of CPUs> are going to do fast path work.

        NVMe IRQs have the following name convention:
             nvme<device index>q<queue index>, e.g. nvme0q7

        :param irq: IRQ number
        :return: True if this IRQ is an IRQ of a FP NVMe queue.
        """
        nvme_irq_re = re.compile(r'(\s|^)nvme\d+q(\d+)(\s|$)')

        # There may be more than an single HW queue bound to the same IRQ. In this case queue names are going to be
        # comma separated
        split_line = self.__irqs2procline[irq].split(",")

        for line in split_line:
            m = nvme_irq_re.search(line)
            if m and 0 < int(m.group(2)) <= multiprocessing.cpu_count():
                return True

        return False

    def __group_disks_info_by_type(self):
        """
        Return a map of tuples ( [<disks>], [<irqs>] ), where "disks" are all disks of the specific type
        and "irqs" are the corresponding IRQs.

        It's promised that every element is "disks" and "irqs" is unique.

        The disk types are 'nvme' and 'non-nvme'
        """
        disks_info_by_type = {}
        nvme_disks = set()
        nvme_irqs = set()
        non_nvme_disks = set()
        non_nvme_irqs = set()
        nvme_disk_name_pattern = re.compile('^nvme')

        for disk, irqs in self.__disk2irqs.items():
            if nvme_disk_name_pattern.search(disk):
                nvme_disks.add(disk)
                for irq in irqs:
                    nvme_irqs.add(irq)
            else:
                non_nvme_disks.add(disk)
                for irq in irqs:
                    non_nvme_irqs.add(irq)

        if not (nvme_disks or non_nvme_disks):
            raise Exception("'disks' tuning was requested but no disks were found")

        nvme_irqs = list(nvme_irqs)

        # There is a known issue with Xen hypervisor that exposes itself on AWS i3 instances where nvme module
        # over-allocates HW queues and uses only queues 1,2,3,..., <up to number of CPUs> for data transfer.
        # On these instances we will distribute only these queues.

        if self.is_aws_i3_non_metal_instance:
            nvme_irqs = list(filter(self.__nvme_fast_path_irq_filter, nvme_irqs))

        # Sort IRQs for easier verification
        nvme_irqs.sort(key=lambda irq_num_str: int(irq_num_str))

        disks_info_by_type[DiskPerfTuner.SupportedDiskTypes.nvme] = (list(nvme_disks), nvme_irqs)
        disks_info_by_type[DiskPerfTuner.SupportedDiskTypes.non_nvme] = ( list(non_nvme_disks), list(non_nvme_irqs) )

        return disks_info_by_type

    def __learn_directories(self):
        return { directory : self.__learn_directory(directory) for directory in self.args.dirs }

    def __learn_directory(self, directory, recur=False):
        """
        Returns a list of disks the given directory is mounted on (there will be more than one if
        the mount point is on the RAID volume)
        """
        if not os.path.exists(directory):
            if not recur:
                perftune_print("{} doesn't exist - skipping".format(directory))

            return []

        try:
            udev_obj = pyudev.Devices.from_device_number(self.__pyudev_ctx, 'block', os.stat(directory).st_dev)
            return self.__get_phys_devices(udev_obj)
        except:
            # handle cases like ecryptfs where the directory is mounted to another directory and not to some block device
            filesystem = run_read_only_command(['df', '-P', directory]).splitlines()[-1].split()[0].strip()
            if not re.search(r'^/dev/', filesystem):
                devs = self.__learn_directory(filesystem, True)
            else:
                raise Exception("Logic error: failed to create a udev device while 'df -P' {} returns a {}".format(directory, filesystem))

            # log error only for the original directory
            if not recur and not devs:
                perftune_print("Can't get a block device for {} - skipping".format(directory))

            return devs

    def __get_phys_devices(self, udev_obj):
        # if device is a virtual device - the underlying physical devices are going to be its slaves
        if re.search(r'virtual', udev_obj.sys_path):
            slaves = os.listdir(os.path.join(udev_obj.sys_path, 'slaves'))
            # If the device is virtual but doesn't have slaves (e.g. as nvm-subsystem virtual devices) handle it
            # as a regular device.
            if slaves:
                return list(itertools.chain.from_iterable([ self.__get_phys_devices(pyudev.Devices.from_device_file(self.__pyudev_ctx, "/dev/{}".format(slave))) for slave in slaves ]))

        # device node is something like /dev/sda1 - we need only the part without /dev/
        return [ re.match(r'/dev/(\S+\d*)', udev_obj.device_node).group(1) ]

    def __learn_irqs(self):
        disk2irqs = {}

        for devices in list(self.__dir2disks.values()) + [ self.args.devs ]:
            for device in devices:
                # There could be that some of the given directories are on the same disk.
                # There is no need to rediscover IRQs of the disk we've already handled.
                if device in disk2irqs.keys():
                    continue

                udev_obj = pyudev.Devices.from_device_file(self.__pyudev_ctx, "/dev/{}".format(device))
                dev_sys_path = udev_obj.sys_path

                # If the device is iSCSI disk, we should skip discovering IRQs
                # since it's virtual device
                if re.search(r'^/sys/devices/platform/host[0-9]+/session[0-9]+', udev_obj.sys_path):
                    disk2irqs[device] = []
                    continue

                # If the device is a virtual NVMe device it's sys file name goes as follows:
                # /sys/devices/virtual/nvme-subsystem/nvme-subsys0/nvme0n1
                #
                # and then there is this symlink:
                # /sys/devices/virtual/nvme-subsystem/nvme-subsys0/nvme0n1/device/nvme0 -> ../../../pci0000:85/0000:85:01.0/0000:87:00.0/nvme/nvme0
                #
                # So, the "main device" is a "nvme\d+" prefix of the actual device name.
                if re.search(r'virtual', udev_obj.sys_path):
                    m = re.match(r'(nvme\d+)\S*', device)
                    if m:
                        dev_sys_path = "{}/device/{}".format(udev_obj.sys_path, m.group(1))

                split_sys_path = list(pathlib.PurePath(pathlib.Path(dev_sys_path).resolve()).parts)

                # first part is always /sys/devices/pciXXX ...
                controller_path_parts = split_sys_path[0:4]

                # ...then there is a chain of one or more "domain:bus:device.function" followed by the storage device enumeration crap
                # e.g. /sys/devices/pci0000:00/0000:00:1f.2/ata2/host1/target1:0:0/1:0:0:0/block/sda/sda3 or
                #      /sys/devices/pci0000:00/0000:00:02.0/0000:02:00.0/host6/target6:2:0/6:2:0:0/block/sda/sda1
                # We want only the path till the last BDF including - it contains the IRQs information.

                patt = re.compile(r"^[0-9ABCDEFabcdef]{4}:[0-9ABCDEFabcdef]{2}:[0-9ABCDEFabcdef]{2}\.[0-9ABCDEFabcdef]$")
                for split_sys_path_branch in split_sys_path[4:]:
                    if patt.search(split_sys_path_branch):
                        controller_path_parts.append(split_sys_path_branch)
                    else:
                        break

                controler_path_str = functools.reduce(lambda x, y : os.path.join(x, y), controller_path_parts)
                disk2irqs[device] = learn_all_irqs_one(controler_path_str, self.__irqs2procline, 'blkif')

        return disk2irqs

    def __get_feature_file(self, dev_node, path_creator):
        """
        Find the closest ancestor with the given feature and return its ('feature file', 'device node') tuple.

        If there isn't such an ancestor - return (None, None) tuple.

        :param dev_node Device node file name, e.g. /dev/sda1
        :param path_creator A functor that creates a feature file name given a device system file name
        """
        # Sanity check
        if dev_node is None or path_creator is None:
            return None, None

        udev = pyudev.Devices.from_device_file(pyudev.Context(), dev_node)
        feature_file = path_creator(udev.sys_path)

        if os.path.exists(feature_file):
            return feature_file, dev_node
        elif udev.parent is not None:
            return self.__get_feature_file(udev.parent.device_node, path_creator)
        else:
            return None, None

    def __tune_one_feature(self, dev_node, path_creator, value, tuned_devs_set):
        """
        Find the closest ancestor that has the given feature, configure it and
        return True.

        If there isn't such ancestor - return False.

        :param dev_node Device node file name, e.g. /dev/sda1
        :param path_creator A functor that creates a feature file name given a device system file name
        """
        feature_file, feature_node = self.__get_feature_file(dev_node, path_creator)

        if feature_file is None:
            return False

        if feature_node not in tuned_devs_set:
            fwriteln_and_log(feature_file, value)
            tuned_devs_set.add(feature_node)

        return True

    def __tune_io_scheduler(self, dev_node, io_scheduler):
        return self.__tune_one_feature(dev_node, lambda p : os.path.join(p, 'queue', 'scheduler'), io_scheduler, self.__io_scheduler_tuned_devs)

    def __tune_nomerges(self, dev_node):
        return self.__tune_one_feature(dev_node, lambda p : os.path.join(p, 'queue', 'nomerges'), self.__nomerges, self.__nomerges_tuned_devs)

    # If write cache configuration is not requested - return True immediately
    def __tune_write_back_cache(self, dev_node):
        if self.__write_cache_config is None:
            return True

        return self.__tune_one_feature(dev_node, lambda p : os.path.join(p, 'queue', 'write_cache'), self.__write_cache_config, self.__write_back_cache_tuned_devs)

    def __get_io_scheduler(self, dev_node):
        """
        Return a supported scheduler that is also present in the required schedulers list (__io_schedulers).

        If there isn't such a supported scheduler - return None.
        """
        feature_file, feature_node = self.__get_feature_file(dev_node, lambda p : os.path.join(p, 'queue', 'scheduler'))

        lines = readlines(feature_file)
        if not lines:
            return None

        # Supported schedulers appear in the config file as a single line as follows:
        #
        # sched1 [sched2] sched3
        #
        # ...with one or more schedulers where currently selected scheduler is the one in brackets.
        #
        # Return the scheduler with the highest priority among those that are supported for the current device.
        supported_schedulers = frozenset([scheduler.lstrip("[").rstrip("]").rstrip("\n") for scheduler in lines[0].split(" ")])
        return next((scheduler for scheduler in self.__io_schedulers if scheduler in supported_schedulers), None)

    def __tune_disk(self, device):
        dev_node = "/dev/{}".format(device)
        io_scheduler = self.__get_io_scheduler(dev_node)

        if not io_scheduler:
            perftune_print("Not setting I/O Scheduler for {} - required schedulers ({}) are not supported".format(device, list(self.__io_schedulers)))
        elif not self.__tune_io_scheduler(dev_node, io_scheduler):
            perftune_print("Not setting I/O Scheduler for {} - feature not present".format(device))

        if not self.__tune_nomerges(dev_node):
            perftune_print("Not setting 'nomerges' for {} - feature not present".format(device))

        if not self.__tune_write_back_cache(dev_node):
                perftune_print("Not setting 'write_cache' for {} - feature not present".format(device))

    def __tune_disks(self, disks):
        for disk in disks:
            self.__tune_disk(disk)

################################################################################
class TuneModes(enum.Enum):
    disks = 0
    net = 1
    system = 2

    @staticmethod
    def names():
        return list(TuneModes.__members__.keys())

# Seastar defaults to allocating 93% of physical memory. The kernel's default allocation for TCP is ~9%. This adds up
# to 102%. Reduce the TCP allocation to 3% to avoid OOM.
default_tcp_mem_fraction = 0.03

argp = argparse.ArgumentParser(description = 'Configure various system parameters in order to improve the seastar application performance.', formatter_class=argparse.RawDescriptionHelpFormatter,
                               epilog=
'''
This script will:

    - Ban relevant IRQs from being moved by irqbalance.
    - Configure various system parameters in /proc/sys.
    - Distribute the IRQs (using SMP affinity configuration) among CPUs according to the configuration mode (see below)
      or an 'irq_cpu_mask' value.

As a result some of the CPUs may be destined to only handle the IRQs and taken out of the CPU set
that should be used to run the seastar application ("compute CPU set").

Modes description:

 sq - set all IRQs of a given NIC to CPU0 and configure RPS
      to spreads NAPIs' handling between other CPUs.

 sq_split - divide all IRQs of a given NIC between CPU0 and its HT siblings and configure RPS
      to spreads NAPIs' handling between other CPUs.

 mq - distribute NIC's IRQs among all CPUs instead of binding
      them all to CPU0. In this mode RPS is always enabled to
      spreads NAPIs' handling between all CPUs.

 If there isn't any mode given script will use a default mode:
    - If number of CPU cores is greater than 16, allocate a single IRQ CPU core for each 16 CPU cores in 'cpu_mask'.
      IRQ cores are going to be allocated evenly on available NUMA nodes according to 'cpu_mask' value.
    - If number of physical CPU cores per Rx HW queue is greater than 4 and less than 16 - use the 'sq-split' mode.
    - Otherwise, if number of hyper-threads per Rx HW queue is greater than 4 - use the 'sq' mode.
    - Otherwise use the 'mq' mode.

Default values:

 --nic NIC       - default: eth0
 --cpu-mask MASK - default: all available cores mask
 --tune-clock    - default: false
''')
argp.add_argument('--mode', choices=PerfTunerBase.SupportedModes.names(), help='configuration mode (deprecated, use --irq-cpu-mask instead)')
argp.add_argument('--nic', action='append', help='network interface name(s), by default uses \'eth0\' (may appear more than once)', dest='nics', default=[])
argp.add_argument('--tune-clock', action='store_true', help='Force tuning of the system clocksource')
argp.add_argument('--get-cpu-mask', action='store_true', help="print the CPU mask to be used for compute")
argp.add_argument('--get-cpu-mask-quiet', action='store_true', help="print the CPU mask to be used for compute, print the zero CPU set if that's what it turns out to be")
argp.add_argument('--get-irq-cpu-mask', action='store_true', help="print the CPU mask to be used for IRQs binding")
argp.add_argument('--verbose', action='store_true', help="be more verbose about operations and their result")
argp.add_argument('--tune', choices=TuneModes.names(), help="components to configure (may be given more than once)", action='append', default=[])
argp.add_argument('--cpu-mask', help="mask of cores to use, by default use all available cores", metavar='MASK')
argp.add_argument('--irq-cpu-mask', help="mask of cores to use for IRQs binding", metavar='MASK')
argp.add_argument('--dir', help="directory to optimize (may appear more than once)", action='append', dest='dirs', default=[])
argp.add_argument('--dev', help="device to optimize (may appear more than once), e.g. sda1", action='append', dest='devs', default=[])
argp.add_argument('--options-file', help="configuration YAML file")
argp.add_argument('--dump-options-file', action='store_true', help="Print the configuration YAML file containing the current configuration")
argp.add_argument('--dry-run', action='store_true', help="Don't take any action, just recommend what to do.")
argp.add_argument('--write-back-cache', help="Enable/Disable \'write back\' write cache mode.", dest="set_write_back")
argp.add_argument('--arfs', help="Enable/Disable aRFS", dest="enable_arfs")
argp.add_argument('--num-rx-queues', help="Set a given number of Rx queues", type=int)
argp.add_argument('--irq-core-auto-detection-ratio', help="Use a given ratio for IRQ mask auto-detection. For "
                                                          "instance, if 8 is given and auto-detection is requested, a "
                                                          "single IRQ CPU core is going to be allocated for every 8 "
                                                          "CPU cores out of available according to a 'cpu_mask' value."
                                                          "Default is 16",
                  type=int, default=16, dest='cores_per_irq_core')
argp.add_argument('--tcp-mem-fraction', default=default_tcp_mem_fraction, type=float, help="Fraction of total memory to allocate for TCP buffers")

def parse_cpu_mask_from_yaml(y, field_name, fname):
    hex_32bit_pattern='0x[0-9a-fA-F]{1,8}'
    mask_pattern = re.compile('^{}((,({})?)*,{})*$'.format(hex_32bit_pattern, hex_32bit_pattern, hex_32bit_pattern))

    if mask_pattern.match(str(y[field_name])):
        return y[field_name]
    else:
        raise Exception("Bad '{}' value in {}: {}".format(field_name, fname, str(y[field_name])))

def extend_and_unique(orig_list, iterable):
    """
    Extend items to a list, and make the list items unique
    """
    assert(isinstance(orig_list, list))
    assert(isinstance(iterable, list))
    orig_list.extend(iterable)
    return list(set(orig_list))

def parse_tri_state_arg(value, arg_name):
    try:
        if value is not None:
            return distutils.util.strtobool(value)
        else:
            return None
    except:
        sys.exit("Invalid {} value: should be boolean but given: {}".format(arg_name, value))

def parse_options_file(prog_args):
    if not prog_args.options_file:
        return

    y = yaml.safe_load(open(prog_args.options_file))
    if y is None:
        return

    if 'mode' in y and not prog_args.mode:
        if not y['mode'] in PerfTunerBase.SupportedModes.names():
            raise Exception("Bad 'mode' value in {}: {}".format(prog_args.options_file, y['mode']))
        prog_args.mode = y['mode']

    if 'nic' in y:
        # Multiple nics was supported by commit a2fc9d72c31b97840bc75ae49dbd6f4b6d394e25
        # `nic' option dumped to config file will be list after this change, but the `nic'
        # option in old config file is still string, which was generated before this change.
        # So here convert the string option to list.
        if not isinstance(y['nic'], list):
            y['nic'] = [y['nic']]
        prog_args.nics = extend_and_unique(prog_args.nics, y['nic'])

    if 'tune_clock' in y and not prog_args.tune_clock:
        prog_args.tune_clock= y['tune_clock']

    if 'tune' in y:
        if set(y['tune']) <= set(TuneModes.names()):
            prog_args.tune = extend_and_unique(prog_args.tune, y['tune'])
        else:
            raise Exception("Bad 'tune' value in {}: {}".format(prog_args.options_file, y['tune']))

    if 'cpu_mask' in y and not prog_args.cpu_mask:
        prog_args.cpu_mask = parse_cpu_mask_from_yaml(y, 'cpu_mask', prog_args.options_file)

    if 'irq_cpu_mask' in y and not prog_args.irq_cpu_mask:
        prog_args.irq_cpu_mask = parse_cpu_mask_from_yaml(y, 'irq_cpu_mask', prog_args.options_file)

    if 'dir' in y:
        prog_args.dirs = extend_and_unique(prog_args.dirs, y['dir'])

    if 'dev' in y:
        prog_args.devs = extend_and_unique(prog_args.devs, y['dev'])

    if 'write_back_cache' in y:
        prog_args.set_write_back = distutils.util.strtobool("{}".format(y['write_back_cache']))

    if 'arfs' in y:
        prog_args.enable_arfs = distutils.util.strtobool("{}".format(y['arfs']))

    if 'num_rx_queues' in y:
        prog_args.num_rx_queues = int(y['num_rx_queues'])

    # prog_options['irq_core_auto_detection_ratio'] = prog_args.cores_per_irq_core
    if 'irq_core_auto_detection_ratio' in y:
        prog_args.cores_per_irq_core = int(y['irq_core_auto_detection_ratio'])

def dump_config(prog_args):
    prog_options = {}

    if prog_args.mode:
        assert prog_args.cpu_mask, "cpu_mask has to always be set. Something is terribly wrong (a bug in perftune.py?)"
        mode = PerfTunerBase.SupportedModes[prog_args.mode]
        prog_options['irq_cpu_mask'] = PerfTunerBase.irqs_cpu_mask_for_mode(mode, prog_args.cpu_mask)

    if prog_args.nics:
        prog_options['nic'] = list(set(prog_args.nics))

    if prog_args.tune_clock:
        prog_options['tune_clock'] = prog_args.tune_clock

    if prog_args.tune:
        prog_options['tune'] = list(set(prog_args.tune))

    if prog_args.cpu_mask:
        prog_options['cpu_mask'] = prog_args.cpu_mask

    if prog_args.irq_cpu_mask:
        prog_options['irq_cpu_mask'] = prog_args.irq_cpu_mask

    if prog_args.dirs:
        prog_options['dir'] = list(set(prog_args.dirs))

    if prog_args.devs:
        prog_options['dev'] = list(set(prog_args.devs))

    if prog_args.set_write_back is not None:
        prog_options['write_back_cache'] = prog_args.set_write_back

    if prog_args.enable_arfs is not None:
        prog_options['arfs'] = prog_args.enable_arfs

    if prog_args.num_rx_queues is not None:
        prog_options['num_rx_queues'] = f"{prog_args.num_rx_queues}"

    prog_options['irq_core_auto_detection_ratio'] = prog_args.cores_per_irq_core

    perftune_print(yaml.dump(prog_options, default_flow_style=False))
################################################################################

args = argp.parse_args()

# Sanity check
args.set_write_back = parse_tri_state_arg(args.set_write_back, "--write-back-cache/write_back_cache")
args.enable_arfs = parse_tri_state_arg(args.enable_arfs, "--arfs/arfs")

dry_run_mode = args.dry_run
parse_options_file(args)

# if nothing needs to be configured - quit
if not args.tune:
    sys.exit("ERROR: At least one tune mode MUST be given.")

# The must be either 'mode' or an explicit 'irq_cpu_mask' given - not both
if args.mode and args.irq_cpu_mask:
    sys.exit("ERROR: Provide either tune mode or IRQs CPU mask - not both.")

# Sanity check
if args.cores_per_irq_core < PerfTunerBase.min_cores_per_irq_core():
    sys.exit(f"ERROR: irq_core_auto_detection_ratio value must be greater or equal than "
             f"{PerfTunerBase.min_cores_per_irq_core()}")

# set default values #####################
if not args.nics:
    args.nics = ['eth0']

if not args.cpu_mask:
    args.cpu_mask = run_hwloc_calc(['all'])
##########################################

# Sanity: irq_cpu_mask should be a subset of cpu_mask
if args.irq_cpu_mask and run_hwloc_calc([args.cpu_mask]) != run_hwloc_calc([args.cpu_mask, args.irq_cpu_mask]):
    sys.exit("ERROR: IRQ CPU mask({}) must be a subset of CPU mask({})".format(args.irq_cpu_mask, args.cpu_mask))

if args.dump_options_file:
    dump_config(args)
    sys.exit(0)

try:
    tuners = []

    if TuneModes.disks.name in args.tune:
        tuners.append(DiskPerfTuner(args))

    if TuneModes.net.name in args.tune:
        tuners.append(NetPerfTuner(args))

    if TuneModes.system.name in args.tune:
        tuners.append(SystemPerfTuner(args))

    if args.get_cpu_mask or args.get_cpu_mask_quiet:
        # Print the compute mask from the first tuner - it's going to be the same in all of them
        perftune_print(tuners[0].compute_cpu_mask)
    elif args.get_irq_cpu_mask:
        perftune_print(tuners[0].irqs_cpu_mask)
    else:
        # Tune the system
        restart_irqbalance(itertools.chain.from_iterable([ tuner.irqs for tuner in tuners ]))

        for tuner in tuners:
            tuner.tune()
except PerfTunerBase.CPUMaskIsZeroException as e:
    # Print a zero CPU set if --get-cpu-mask-quiet was requested.
    if args.get_cpu_mask_quiet:
        perftune_print("0x0")
    else:
        sys.exit("ERROR: {}. Your system can't be tuned until the issue is fixed.".format(e))
except Exception as e:
    sys.exit("ERROR: {}. Your system can't be tuned until the issue is fixed.".format(e))




#!/usr/bin/env python3
#
# This file is open source software, licensed to you under the terms
# of the Apache License, Version 2.0 (the "License").  See the NOTICE file
# distributed with this work for additional information regarding copyright
# ownership.  You may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import argparse
import subprocess
import seastar_cmake

if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Seastar test runner")
    parser.add_argument('--fast',  action="store_true", help="Run only fast tests")
    parser.add_argument('--name',  action="store", help="Run only test whose name contains given string")
    parser.add_argument('--mode', choices=seastar_cmake.SUPPORTED_MODES, help="Run only tests for given build mode")
    parser.add_argument('--build-root', action='store', default=seastar_cmake.DEFAULT_BUILD_ROOT, type=str,
                        help="The name of the build root build directoy: "
                        "using a different name allows multiple configurations to co-exist in the same repository")
    parser.add_argument('--timeout', action="store",default="300",type=int, help="timeout value for test execution")
    parser.add_argument('--jenkins', action="store",help="jenkins output file prefix")
    parser.add_argument('--smp', '-c', action="store",default='2',type=int,help="Number of threads for multi-core tests")
    parser.add_argument('--verbose', '-v', action = 'store_true', default = False,
                        help = 'Verbose reporting')
    parser.add_argument('--offline', action="store_true", default = False,
                        help="Disable tests accessing internet")
    parser.add_argument('ctest_forward', nargs='*', help="These parameters will be passed directly to ctest")
    args = parser.parse_args()

    MODES = [args.mode] if args.mode else seastar_cmake.SUPPORTED_MODES

    def run_tests(mode):
        BUILD_PATH = seastar_cmake.build_path(mode, args.build_root)

        # For convenience.
        tr = seastar_cmake.translate_arg

        TRANSLATED_CMAKE_ARGS = [
            tr(args.timeout, 'TEST_TIMEOUT'),
            tr(args.fast, 'EXECUTE_ONLY_FAST_TESTS'),
            tr(args.smp, 'UNIT_TEST_SMP'),
            tr(not args.offline, 'ENABLE_TESTS_ACCESSING_INTERNET'),
            tr(args.jenkins, 'JENKINS', value_when_none=''),
        ]

        # Modify the existing build by pointing to the build directory.
        CMAKE_ARGS = ['cmake', BUILD_PATH] + TRANSLATED_CMAKE_ARGS
        print(CMAKE_ARGS)
        subprocess.check_call(CMAKE_ARGS, shell=False, cwd=seastar_cmake.ROOT_PATH)

        TRANSLATED_CTEST_ARGS = ['--output-on-failure']
        if args.verbose:
            TRANSLATED_CTEST_ARGS += ['--verbose']
        if args.name:
            TRANSLATED_CTEST_ARGS += ['-R', args.name]

        CTEST_ARGS = ['ctest', BUILD_PATH] + TRANSLATED_CTEST_ARGS + args.ctest_forward
        print(CTEST_ARGS)
        subprocess.check_call(CTEST_ARGS, shell=False, cwd=BUILD_PATH)

    for mode in MODES:
        run_tests(mode)



#!/usr/bin/env python3

# C++ Code generation utility from Swagger definitions.
# This utility support Both the swagger 1.2 format
#    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/1.2.md
# And the 2.0 format
#    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md
#
# Swagger 2.0 is not only different in its structure (apis have moved, and
# models are now under definitions) It also moved from multiple file structure
# to a single file.
# To keep the multiple file support, each group of APIs will be placed in a single file
# Each group can have a .def.json file with its definitions (What used to be models)
# Because the APIs in definitions are snippets, they are not legal json objects
# and need to be formated as such so that a json parser would work.

import json
import sys
import re
import glob
import argparse
import os
import textwrap
from string import Template

parser = argparse.ArgumentParser(description="""Generate C++ class for json
handling from swagger definition""")

parser.add_argument('--outdir', help='the output directory', default='autogen')
parser.add_argument('-o', help='Output file', default='')
parser.add_argument('-f', help='input file', default='api-java.json')
parser.add_argument('-ns', help="""namespace when set struct will be created
under the namespace""", default='')
parser.add_argument('-jsoninc', help='relative path to the jsaon include',
                    default='json/')
parser.add_argument('-jsonns', help='set the json namespace', default='json')
parser.add_argument('-indir', help="""when set all json file in the given
directory will be parsed, do not use with -f""", default='')
parser.add_argument('-debug', help='debug level 0 -quite,1-error,2-verbose',
                    default='1', type=int)
parser.add_argument('-combined', help='set the name of the combined file',
                    default='autogen/pathautogen.ee')
parser.add_argument('--create-cc', dest='create_cc', action='store_true', default=False,
                    help='Put global variables in a .cc file')
config = parser.parse_args()


valid_vars = {'string': 'sstring', 'int': 'int', 'double': 'double',
             'float': 'float', 'long': 'long', 'boolean': 'bool', 'char': 'char',
             'datetime': 'json::date_time'}

current_file = ''

spacing = "    "
def getitem(d, key, name):
    item = d.get(key)
    if item is None:
        raise Exception(f"'{key}' not found in {name}")
    return item

def fprint(f, *args):
    for arg in args:
        f.write(arg)

def fprintln(f, *args):
    for arg in args:
        f.write(arg)
    f.write('\n')


def open_namespace(f, ns=config.ns):
    fprintln(f, "namespace ", ns , ' {\n')


def close_namespace(f):
    fprintln(f, '}')


def add_include(f, includes):
    for include in includes:
        fprintln(f, '#include ', include)
    fprintln(f, "")

def trace_verbose(*params):
    if config.debug > 1:
        print(''.join(params))


def trace_err(*params):
    if config.debug > 0:
        print(current_file + ':' + ''.join(params))


def valid_type(param):
    if param in valid_vars:
        return valid_vars[param]
    trace_err("Type [", param, "] not defined")
    return param

# Map from json list types to the C++ implementing type
LIST_TO_IMPL = {"array": "json_list", "chunked_array": "json_chunked_list"}

def is_array_type(type: str):
    return type in LIST_TO_IMPL

def type_change(param: str, member):
    if is_array_type(param):
        if "items" not in member:
            trace_err("array without item declaration in ", param)
            return ""
        item = member["items"]
        if "type" in item:
            t = item["type"]
        elif "$ref" in item:
            t = item["$ref"]
        else:
            trace_err("array items with no type or ref declaration ", param)
            return ""
        return LIST_TO_IMPL[param] + "< " + valid_type(t) + " >"

    return "json_element< " + valid_type(param) + " >"



def print_ind_comment(f, ind, comment):
    fprintln(f, ind, "/**")
    fprintln(f, ind, " * ", comment)
    fprintln(f, ind, " */")

def print_comment(f, comment):
    print_ind_comment(f, spacing, comment)

def print_copyrights(f):
    fprintln(f, "/*")
    fprintln(f, "* Copyright (C) 2014 Cloudius Systems, Ltd.")
    fprintln(f, "*")
    fprintln(f, "* This work is open source software, licensed under the",
           " terms of the")
    fprintln(f, "* BSD license as described in the LICENSE f in the top-",
           "level directory.")
    fprintln(f, "*")
    fprintln(f, "*  This is an Auto-Generated-code  ")
    fprintln(f, "*  Changes you do in this file will be erased on next",
           " code generation")
    fprintln(f, "*/\n")


def print_h_file_headers(f, name):
    print_copyrights(f)
    fprintln(f, "#ifndef __JSON_AUTO_GENERATED_" + name)
    fprintln(f, "#define __JSON_AUTO_GENERATED_" + name + "\n")


def clean_param(param):
    match = re.match(r"^\{\s*([^\}]+)\s*}", param)
    if match:
        return [match.group(1), False]
    return [param, True]


def get_parameter_by_name(obj, name):
    for p in obj["parameters"]:
        if p["name"] == name:
            return p
    trace_err ("No Parameter declaration found for ", name)


def clear_path_ending(path):
    if not path or path[-1] != '/':
        return path
    return path[0:-1]


class Parameter:
    '''represents a parameter

    TODO: return an instance of Parameter from get_parameter_by_name()
    '''
    def __init__(self, definition):
        self.definition = definition

    @property
    def name(self):
        return self.definition['name']

    @property
    def is_required(self):
        # check if a parameter is query required.
        # It will return true if the required flag is set
        # and if it is a query parameter, both swagger 1.2 'paramType' and
        # swagger 2.0 'in' attributes are supported
        if "required" not in self.definition:
            return False
        if not self.definition["required"]:
            return False
        if self.definition.get("paramType") == "query":
            return True
        if self.definition.get("in") == "query":
            return True
        return False

    @property
    def enum(self):
        return self.definition.get('enum')


def add_path(f, path, details):
    if "summary" in details:
        print_comment(f, details["summary"])
    param_starts = path.find("{")
    if param_starts >= 0:
        path_reminder = path[param_starts:]
        vals = path.split("/")
        vals.reverse()
        fprintln(f, spacing, 'path_description::add_path("', clear_path_ending(vals.pop()),
           '",', details["method"], ',"', details["nickname"], '")')
        while vals:
            param, is_url = clean_param(vals.pop())
            if is_url:
                fprintln(f, spacing, '  ->pushurl("', param, '")')
            else:
                param_type = get_parameter_by_name(details, param)
                if ("allowMultiple" in param_type and
                    param_type["allowMultiple"]):
                    fprintln(f, spacing, '  ->pushparam("', param, '",true)')
                else:
                    fprintln(f, spacing, '  ->pushparam("', param, '")')
    else:
        fprintln(f, spacing, 'path_description::add_path("', clear_path_ending(path), '",',
           details["method"], ',"', details["nickname"], '")')
    if "parameters" in details:
        for param_definition in details["parameters"]:
            param = Parameter(param_definition)
            if param.is_required:
                fprintln(f, spacing, '  ->pushmandatory_param("', param.name, '")')
    fprintln(f, spacing, ";")


def not_first():
    '''
    Returns True when gets called for the first time, False otherwise

    used as the predicate parameter of textwrap.indent(), so the first
    line is not indented. this helps to preserve the Python code's logical
    indention in the template, and allows us to put something like::

      blah = textwrap.indent("""\
           foo bar
               blah blah
           foo bar
    """
    '''
    _is_first = True

    def should_indent(_):
        nonlocal _is_first
        first = _is_first
        _is_first = False
        return not first
    return should_indent


def generate_code_from_enum(nickname, type_name, enums):
    def indent_body(s, level):
        return textwrap.indent(s, level * '    ', not_first())

    enum_list = ',\n'.join(enums + ['NUM_ITEMS'])
    decl = Template('''\
    namespace ns_$nickname {
        enum class $type_name {
            $enum_list
        };
        $type_name str2$type_name(const sstring& str);
   }
   ''').substitute(nickname=nickname,
                   type_name=type_name,
                   enum_list=indent_body(enum_list, 3))

    name_list = ',\n'.join(f'"{enum}"' for enum in enums)
    parse_func = Template('''\
    $type_name str2$type_name(const sstring& str) {
        static const std::string_view arr[] = {
            $name_list
        };
        int i;
        for (i = 0; i < $num_enums; i++) {
            if (arr[i] == str) {
                return ($type_name)i;
            }
        }
        return ($type_name)i;
    }
    ''').substitute(type_name=type_name,
                    name_list=indent_body(name_list, 3),
                    num_enums=len(enums))

    return decl, parse_func


def add_operation(hfile, ccfile, path, oper):
    if "summary" in oper:
        print_ind_comment(hfile, '', oper["summary"])

    param_starts = path.find("{")
    base_url = path
    vals = []
    if param_starts >= 0:
        vals = path[param_starts:].split("/")
        vals.reverse()
        base_url = path[:param_starts]

    nickname = getitem(oper, "nickname", oper)
    if config.create_cc:
        fprintln(hfile, f'extern const path_description {nickname};')
        maybe_static = ''
    else:
        maybe_static = 'static '
    normalized_path = clear_path_ending(base_url)
    http_method = oper["method"]
    fprintln(ccfile, f'{maybe_static}const path_description {nickname}("{normalized_path}",{http_method},"{nickname}",')
    fprint(ccfile, '{')
    first = True
    while vals:
        path_param, is_url = clean_param(vals.pop())
        if path_param == "":
            continue
        if first:
            first = False
        else:
            fprint(ccfile, "\n,")
        if is_url:
            path_param = f"/{path_param}"
            component_type = 'FIXED_STRING'
        elif get_parameter_by_name(oper, path_param).get("allowMultiple",
                                                         False):
            component_type = 'PARAM_UNTIL_END_OF_PATH'
        else:
            component_type = 'PARAM'
        fprint(ccfile, f'{{"{path_param}", path_description::url_component_type::{component_type}}}')
    fprint(ccfile, '}')
    fprint(ccfile, ',{')
    enum_definitions = ""
    if "enum" in oper:
        enum_wrapper = create_enum_wrapper(nickname, "return_type", oper["enum"])
        enum_definitions = Template('''
namespace ns_$nickname {
$enum_wrapper
}
''').substitute(nickname=nickname, enum_wrapper=enum_wrapper.rstrip())
    funcs = ""
    required_query_params = []
    for param in oper.get("parameters", []):
        query_param = Parameter(param)
        if query_param.is_required:
            required_query_params.append(query_param)
        if query_param.enum is not None:
            enum_decl, parse_func = generate_code_from_enum(nickname,
                                                            query_param.name,
                                                            query_param.enum)
            enum_definitions += enum_decl
            funcs += parse_func
    fprint(ccfile, '\n,'.join(f'"{param.name}"' for param in required_query_params))
    fprintln(ccfile, '});')
    fprintln(hfile, enum_definitions)
    open_namespace(ccfile, f'ns_{nickname}')
    fprintln(ccfile, funcs)
    close_namespace(ccfile)


def get_base_name(param):
    return os.path.basename(param)


def is_model_valid(name, model):
    if name in valid_vars:
        return ""
    properties = getitem(model[name], "properties", name)
    for var in properties:
        type = getitem(properties[var], "type", name + ":" + var)
        if is_array_type(type):
            items = getitem(properties[var], "items", name + ":" + var)
            try:
                type = getitem(items, "type", name + ":" + var + ":items")
            except Exception as e:
                try:
                    type = getitem(items, "$ref", name + ":" + var + ":items")
                except:
                    raise e
        if type not in valid_vars:
            if type not in model:
                raise Exception("Unknown type '" + type + "' in Model '" + name + "'")
            return type
    valid_vars[name] = name
    return ""

def resolve_model_order(data):
    res = []
    models = set()
    for model_name in data:
        visited = set(model_name)
        missing = is_model_valid(model_name, data)
        resolved = missing == ''
        if not resolved:
            stack = [model_name]
            while not resolved:
                if missing in visited:
                    raise Exception("Cyclic dependency found: " + missing)
                missing_depends = is_model_valid(missing, data)
                if missing_depends == '':
                    if missing not in models:
                        res.append(missing)
                        models.add(missing)
                    resolved = len(stack) == 0
                    if not resolved:
                        missing = stack.pop()
                else:
                    stack.append(missing)
                    missing = missing_depends
        elif model_name not in models:
            res.append(model_name)
            models.add(model_name)
    return res


def create_enum_wrapper(model_name, name, values):
    enum_name = f"{model_name}_{name}"
    wrapper = name + "_wrapper"

    def indent_body(s, level):
        return textwrap.indent(s, level * '    ', not_first())

    case_clauses = "\n".join(
        Template('''case $enum_name::$enum_entry: return "\\"$enum_entry\\"";''').substitute(
            enum_name=enum_name, enum_entry=enum_entry) for enum_entry in values)
    res = Template("""\
    virtual std::string to_json() const {
        switch(v) {
        $case_clauses
        default: return "\\"Unknown\\"";
        }
    }""").substitute(wrapper=wrapper,
                     case_clauses=indent_body(case_clauses, 2))

    case_clauses = "\n".join(
        Template("""case T::$enum_entry: v = $enum_name::$enum_entry; break;""").substitute(
            enum_name=enum_name, enum_entry=enum_entry) for enum_entry in values)
    res += Template("""
    template<class T>
    $wrapper(const T& _v) {
        switch(_v) {
        $case_clauses
        default: v = $enum_name::NUM_ITEMS;
        }
    }""").substitute(wrapper=wrapper,
                     case_clauses=indent_body(case_clauses, 2),
                     enum_name=enum_name)

    case_clauses = "\n".join(
        Template('''case $enum_name::$enum_entry: return T::$enum_entry;''').substitute(
            enum_name=enum_name, enum_entry=enum_entry) for enum_entry in values)
    res += Template("""
    template<class T>
    operator T() const {
        switch(v) {
        $case_clauses
        default: return T::$value;
        }
    }""").substitute(case_clauses=indent_body(case_clauses, 2),
                     enum_name=enum_name,
                     value=values[0])

    res += Template("""
    typedef typename std::underlying_type<$enum_name>::type pos_type;
    $wrapper& operator++() {
        v = static_cast<$enum_name>(static_cast<pos_type>(v) + 1);
        return *this;
    }
    $wrapper & operator++(int) {
        return ++(*this);
    }
    bool operator==(const  $wrapper& c) const {
        return v == c.v;
    }
    bool operator!=(const $wrapper& c) const {
        return v != c.v;
    }
    bool operator<=(const $wrapper& c) const {
        return static_cast<pos_type>(v) <= static_cast<pos_type>(c.v);
    }
    std::make_signed_t<pos_type> operator-(const $wrapper& c) const {
        return static_cast<pos_type>(v) - static_cast<pos_type>(c.v);
    }
    static $wrapper begin() {
        return $wrapper($enum_name::$value);
    }
    static $wrapper end() {
        return $wrapper($enum_name::NUM_ITEMS);
    }
    static auto /* iota_view */ all_items() {
        return std::ranges::iota_view<$wrapper, $wrapper>(begin(), end());
    }
    $enum_name v;""").substitute(enum_name=enum_name,
                                 wrapper=wrapper,
                                 value=values[0])
    enum_wrapper = Template("""
    enum class $enum_name { $enumerators, NUM_ITEMS };
    struct $wrapper : public json::jsonable {
        $wrapper() = default;
        $body
    };
    """).substitute(enum_name=enum_name,
                    enumerators=", ".join(values),
                    wrapper=wrapper,
                    body=indent_body(res, 1).lstrip())
    return enum_wrapper.lstrip('\n')


def to_operation(opr, data):
    data["method"] = opr.upper()
    data["nickname"] = data["operationId"]
    return data

def to_path(path, data):
    data["operations"] = [to_operation(k, data[k]) for k in data]
    data["path"] = path

    return data

def create_h_file(data, hfile_name, api_name, init_method, base_api):
    if config.o != '':
        final_hfile_name = config.o
    else:
        final_hfile_name = config.outdir + "/" + hfile_name
    hfile = open(final_hfile_name, "w")

    if config.create_cc:
        ccfile = open(final_hfile_name.rstrip('.hh') + ".cc", "w")
        add_include(ccfile, ['"{}"'.format(final_hfile_name)])
        open_namespace(ccfile, "seastar")
        open_namespace(ccfile, "httpd")
        open_namespace(ccfile, api_name)
    else:
        ccfile = hfile
    print_h_file_headers(hfile, api_name)
    add_include(hfile, ['<seastar/core/sstring.hh>',
                        '<seastar/json/json_elements.hh>',
                        '<seastar/http/json_path.hh>'])

    add_include(hfile, ['<iostream>', '<ranges>'])
    open_namespace(hfile, "seastar")
    open_namespace(hfile, "httpd")
    open_namespace(hfile, api_name)

    def indent(s):
        return textwrap.indent(s.rstrip(), '        ', not_first())

    if "models" in data:
        models_order = resolve_model_order(data["models"])
        for model_name in models_order:
            model = data["models"][model_name]
            if 'description' in model:
                print_ind_comment(hfile, "", model["description"])
            fprintln(hfile, "struct ", model_name, " : public json::json_base {")
            member_init = ''
            member_assignment = ''
            member_copy = ''
            for member_name in model["properties"]:
                member = model["properties"][member_name]
                if "description" in member:
                    print_comment(hfile, member["description"])
                if "enum" in member:
                    fprintln(hfile, create_enum_wrapper(model_name, member_name, member["enum"]))
                    fprintln(hfile, f"    {config.jsonns}::json_element<{member_name}_wrapper> {member_name};\n")
                else:
                    type_name = type_change(member["type"], member)
                    fprintln(hfile, f"    {config.jsonns}::{type_name} {member_name};\n")
                member_init += f'add(&{member_name}, "{member_name}");\n'
                member_assignment += f'{member_name} = e.{member_name};\n'
                member_copy += f'e.{member_name} = {member_name} ;\n'

            functions = Template('''
    void register_params() {
        $member_init
    }
    $model_name() {
        register_params();
    }
    $model_name(const $model_name& e) {
        register_params();
        $member_assignment
    }
    template<class T>
    $model_name& operator=(const T& e) {
        $member_assignment
        return *this;
    }
    $model_name& operator=(const $model_name& e) {
        $member_assignment
        return *this;
    }
    template<class T>
    $model_name& update(T& e) {
        $member_copy
        return *this;
    }''').substitute(model_name=model_name,
                     member_init=indent(member_init),
                     member_assignment=indent(member_assignment),
                     member_copy=indent(member_copy))
            fprintln(hfile, functions.lstrip('\n'))
            fprintln(hfile, "};\n\n")

 #   print_ind_comment(hfile, "", "Initialize the path")
#    fprintln(hfile, init_method + "(const std::string& description);")
    fprintln(hfile, 'static const sstring name = "', base_api, '";')
    for item in data["apis"]:
        path = item["path"]
        for oper in item.get("operations", []):
            add_operation(hfile, ccfile, path, oper)

    close_namespace(hfile)
    close_namespace(hfile)
    close_namespace(hfile)
    if config.create_cc:
        close_namespace(ccfile)
        close_namespace(ccfile)
        close_namespace(ccfile)

    hfile.write("#endif //__JSON_AUTO_GENERATED_HEADERS\n")
    hfile.close()

def remove_leading_comma(data):
    return re.sub(r'^\s*,', '', data)

def format_as_json_object(data):
    return "{" + remove_leading_comma(data) + "}"

def check_for_models(data, param):
    model_name = param.replace(".json", ".def.json")
    if not os.path.isfile(model_name):
        return
    try:
        with open(model_name) as myfile:
            json_data = myfile.read()
            def_data = json.loads(format_as_json_object(json_data))
            data["models"] = def_data
    except Exception as e:
        type, value, tb = sys.exc_info()
        print("Bad formatted JSON definition file '" + model_name + "' error ", value.message)
        sys.exit(-1)

def set_apis(data):
    return {"apis": [to_path(p, data[p]) for p in data]}

def parse_file(param, combined):
    global current_file
    trace_verbose("parsing ", param, " file")
    with open(param) as myfile:
        json_data = myfile.read()
    try:
        data = json.loads(json_data)
    except Exception as e:
        try:
            # the passed data is not a valid json, so maybe its a swagger 2.0
            # snippet, format it as json and try again
            # set_apis and check_for_models will create an object with a similiar format
            # to a swagger 1.2 so the code generation would work
            data = set_apis(json.loads(format_as_json_object(json_data)))
            check_for_models(data, param)
        except:
            # The problem is with the file,
            # just report the error and exit.
            type, value, tb = sys.exc_info()
            print("Bad formatted JSON file '" + param + "' error ", value.message)
            sys.exit(-1)
    try:
        base_file_name = get_base_name(param)
        current_file = base_file_name
        hfile_name = base_file_name + ".hh"
        api_name = base_file_name.replace('.', '_')
        base_api = base_file_name.replace('.json', '')
        init_method = "void " + api_name + "_init_path"
        trace_verbose("creating ", hfile_name)
        if (combined):
            fprintln(combined, '#include "', base_file_name, ".cc", '"')
        create_h_file(data, hfile_name, api_name, init_method, base_api)
    except Exception as e:
        print("Error while parsing JSON file '" + param + "' error " + str(e))
        sys.exit(-1)

if "indir" in config and config.indir != '':
    combined = open(config.combined, "w")
    for f in glob.glob(os.path.join(config.indir, "*.json")):
        parse_file(f, combined)
else:
    parse_file(config.f, None)



#!/usr/bin/env python3
#
# This file is open source software, licensed to you under the terms
# of the Apache License, Version 2.0 (the "License").  See the NOTICE file
# distributed with this work for additional information regarding copyright
# ownership.  You may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
#
# Copyright (C) 2024 Scylladb, Ltd.
#

import argparse
import math
import json
import re
import subprocess
import sys
import time
import unittest
import urllib.request
import urllib.parse
import yaml

from typing import Optional
from collections import namedtuple


class Exposition:
    @classmethod
    def from_hist(cls,
                  name: str,
                  hist: list[tuple[float, int]],
                  sum_: int,
                  count: int) -> 'Exposition':
        # ignore these values, we might need to verify them in future
        _, _ = sum_, count
        buckets = (cls.value_to_bucket(le - 1) for le, _ in hist)
        deltas = []
        last_n = 0
        for _, n in hist:
            delta = n - last_n
            last_n = n
            deltas.append(delta)
        return cls(name, dict(zip(buckets, deltas)), {})

    @staticmethod
    def value_to_bucket(value):
        low = 2 ** math.floor(math.log(value, 2))
        high = 2 * low
        dif = (high - low) / 4
        return low + dif * math.floor((value - low) / dif)

    @staticmethod
    def _values_to_histogram(values):
        hist = {}
        for val in values:
            bucket = Exposition.value_to_bucket(val)
            if bucket in hist:
                hist[bucket] += 1
            else:
                hist[bucket] = 1
        return hist

    @classmethod
    def from_conf(cls,
                  name: str,
                  type_: str,
                  values: list[str],
                  labels: dict[str, str]) -> 'Exposition':
        if type_ in ('gauge', 'counter'):
            assert len(values) == 1
            return cls(name, float(values[0]), labels)
        if type_ == 'histogram':
            hist = cls._values_to_histogram(float(v) for v in values)
            return cls(name, hist, {})
        raise NotImplementedError(f'unsupported type: {type_}')

    def __init__(self,
                 name: str,
                 value: int | list[tuple[float, int]],
                 labels: dict[str, str]) -> None:
        self.name = name
        self.value = value
        self.labels = labels

    def __repr__(self):
        return f"{self.name=}, {self.value=}, {self.labels=}"

    def __eq__(self, other):
        if not isinstance(other, Exposition):
            return False
        return self.value == other.value


class Metrics:
    prefix = 'seastar'
    group = 'test_group'
    # parse lines like:
    # rest_api_scheduler_queue_length{group="main",shard="0"} 0.000000
    # where:
    #   - "rest_api" is the prometheus prefix
    #   - "scheduler" is the metric group name
    #   - "queue_length" is the name of the metric
    #   - the kv pairs in "{}" are labels"
    #   - "0.000000" is the value of the metric
    # this format is compatible with
    # https://github.com/prometheus/docs/blob/main/content/docs/instrumenting/exposition_formats.md
    # NOTE: scylla does not include timestamp in the exported metrics
    pattern = re.compile(r'''(?P<metric_name>\w+)   # rest_api_scheduler_queue_length
                             \{(?P<labels>[^\}]*)\} # {group="main",shard="0"}
                             \s+                    # <space>
                             (?P<value>[^\s]+)      # 0.000000''', re.X)

    def __init__(self, lines: list[str]) -> None:
        self.lines: list[str] = lines

    @classmethod
    def full_name(cls, name: str) -> str:
        '''return the full name of a metrics
        '''
        return f'{cls.group}_{name}'

    @staticmethod
    def _parse_labels(s: str) -> dict[str, str]:
        return dict(name_value.split('=', 1) for name_value in s.split(','))

    def get(self,
            name: Optional[str] = None,
            labels: Optional[dict[str, str]] = None) -> list[Exposition]:
        '''Return all expositions matching the given name and labels
        '''
        full_name = None
        if name is not None:
            full_name = f'{self.prefix}_{self.group}_{name}'
        metric_type = None

        # for histogram and summary as they are represented with multiple lines
        hist_name = ''
        hist_buckets = []
        hist_sum = 0
        hist_count = 0

        for line in self.lines:
            if not line:
                continue
            if line.startswith('# HELP'):
                continue
            if line.startswith('# TYPE'):
                _, _, type_metric_name, metric_type = line.split()
                if hist_buckets:
                    yield Exposition.from_hist(hist_name,
                                               hist_buckets,
                                               hist_sum,
                                               hist_count)
                    hist_buckets = []
                if metric_type in ('histogram', 'summary'):
                    hist_name = type_metric_name
                continue
            matched = self.pattern.match(line)
            assert matched, f'malformed metric line: {line}'

            value_metric_name = matched.group('metric_name')
            if full_name and not value_metric_name.startswith(full_name):
                continue

            metric_labels = self._parse_labels(matched.group('labels'))
            if labels is not None and metric_labels != labels:
                continue

            metric_value = float(matched.group('value'))
            if metric_type == 'histogram':
                if value_metric_name == f'{type_metric_name}_bucket':
                    last_value = 0
                    if hist_buckets:
                        last_value = hist_buckets[-1][1]
                    if metric_value - last_value != 0:
                        le = metric_labels['le'].strip('"')
                        hist_buckets.append((float(le), metric_value))
                elif value_metric_name == f'{type_metric_name}_sum':
                    hist_sum = metric_value
                elif value_metric_name == f'{type_metric_name}_count':
                    hist_count = metric_value
                else:
                    raise RuntimeError(f'unknown histogram value: {line}')
            elif metric_type == 'summary':
                raise NotImplementedError('unsupported type: summary')
            else:
                yield Exposition(type_metric_name,
                                 metric_value,
                                 metric_labels)
        if hist_buckets:
            yield Exposition.from_hist(hist_name,
                                       hist_buckets,
                                       hist_sum,
                                       hist_count)

    def get_help(self, name: str) -> Optional[str]:
        full_name = f'{self.prefix}_{self.group}_{name}'
        header = f'# HELP {full_name}'
        for line in self.lines:
            if line.startswith(header):
                tokens = line.split(maxsplit=3)
                return tokens[-1]
        return None


class TestPrometheus(unittest.TestCase):
    exporter_path = None
    exporter_process = None
    exporter_config = None
    port = 10001
    prometheus = None
    prometheus_scrape_interval = 15

    @classmethod
    def setUpClass(cls) -> None:
        args = [cls.exporter_path,
                '--port', f'{cls.port}',
                '--conf', cls.exporter_config,
                '--smp=2']
        cls.exporter_process = subprocess.Popen(args,
                                                stdout=subprocess.PIPE,
                                                stderr=subprocess.DEVNULL,
                                                bufsize=0, text=True)
        # wait until the server is ready for serve
        cls.exporter_process.stdout.readline()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.exporter_process.terminate()

    @classmethod
    def _get_metrics(cls,
                     name: Optional[str] = None,
                     labels: Optional[dict[str, str]] = None,
                     with_help: bool = True,
                     aggregate: bool = True) -> Metrics:
        query: dict[str, str] = {}
        if name is not None:
            query['__name__'] = name
        if labels is not None:
            query.update(labels)
        if not with_help:
            query['__help__'] = 'false'
        if not aggregate:
            query['__aggregate__'] = 'false'
        params = urllib.parse.urlencode(query)
        host = 'localhost'
        url = f'http://{host}:{cls.port}/metrics?{params}'
        with urllib.request.urlopen(url) as f:
            body = f.read().decode('utf-8')
            return Metrics(body.rstrip().split('\n'))

    def test_filtering_by_label_sans_aggregation(self) -> None:
        labels = {'private': '1'}
        metrics = self._get_metrics(labels=labels)
        actual_values = list(metrics.get())
        expected_values = []
        with open(self.exporter_config, encoding='utf-8') as f:
            config = yaml.safe_load(f)
        for metric in config['metrics']:
            name = metric['name']
            metric_name = f'{Metrics.prefix}_{Metrics.group}_{name}'
            metric_labels = metric['labels']
            if metric_labels != labels:
                continue
            e = Exposition.from_conf(metric_name,
                                     metric['type'],
                                     metric['values'],
                                     metric_labels)
            expected_values.append(e)
        self.assertCountEqual(actual_values, expected_values)

    def test_filtering_by_label_with_aggregation(self) -> None:
        TestCase = namedtuple('TestCase', ['label', 'regex', 'found'])
        label = 'private'
        tests = [
            TestCase(label=label, regex='dne', found=0),
            TestCase(label=label, regex='404', found=0),
            TestCase(label=label, regex='2', found=1),
            # aggregated
            TestCase(label=label, regex='2|3', found=1),
        ]
        for test in tests:
            with self.subTest(regex=test.regex, found=test.found):
                metrics = self._get_metrics(labels={test.label: test.regex})
                values = list(metrics.get())
                self.assertEqual(len(values), test.found)

    def test_aggregated(self) -> None:
        name = 'counter_1'
        # see also rest_api_httpd.cc::aggregate_by_name
        TestCase = namedtuple('TestCase', ['aggregate', 'expected_values'])
        tests = [
            TestCase(aggregate=False, expected_values=[1, 2]),
            TestCase(aggregate=True, expected_values=[3])
        ]
        for test in tests:
            with self.subTest(aggregate=test.aggregate,
                              values=test.expected_values):
                metrics = self._get_metrics(Metrics.full_name(name), aggregate=test.aggregate)
                expositions = metrics.get(name)
                actual_values = sorted(e.value for e in expositions)
                self.assertEqual(actual_values, test.expected_values)

    def test_help(self) -> None:
        name = 'counter_1'
        tests = [True, False]
        for with_help in tests:
            with self.subTest(with_help=with_help):
                metrics = self._get_metrics(Metrics.full_name(name), with_help=with_help)
                msg = metrics.get_help(name)
                if with_help:
                    self.assertIsNotNone(msg)
                else:
                    self.assertIsNone(msg)

    @staticmethod
    def _from_native_histogram(values) -> dict[float, float]:
        results = {}
        for v in values:
            bucket = Exposition.value_to_bucket(float(v[2]) - 1)
            results[bucket] = float(v[3])
        return results

    @staticmethod
    def _query_prometheus(host: str, query: str, type_: str) -> float | dict[float, float]:
        url = f'http://{host}/api/v1/query?query={query}'
        headers = {"Accept": "application/json"}
        req = urllib.request.Request(url, headers=headers)
        with urllib.request.urlopen(req) as f:
            results = json.load(f)["data"]["result"][0]
            if type_ == 'histogram':
                buckets = results["histogram"][1]["buckets"]
                return TestPrometheus._from_native_histogram(buckets)
            return float(results["value"][1])

    def test_protobuf(self) -> None:
        if self.prometheus is None:
            self.skipTest("prometheus is not configured")

        # Prometheus does not allow us to push metrics to it, neither
        # can we force it to scrape an exporter, so we have to wait
        # until prometheus scrapes the server
        time.sleep(self.prometheus_scrape_interval + 1)
        with open(self.exporter_config, encoding='utf-8') as f:
            config = yaml.safe_load(f)

        labels = {'private': '1'}
        for metric in config['metrics']:
            name = metric['name']
            metric_name = f'{Metrics.prefix}_{Metrics.group}_{name}'
            metric_labels = metric['labels']
            if metric_labels != labels:
                continue
            metric_type = metric['type']
            metric_value = metric['values']
            e = Exposition.from_conf(metric_name,
                                     metric_type,
                                     metric_value,
                                     metric_labels)
            res = self._query_prometheus(self.prometheus,
                                         metric_name,
                                         metric_type)
            self.assertEqual(res, e.value)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--exporter',
                        required=True,
                        help='Path to the exporter executable')
    parser.add_argument('--config',
                        required=True,
                        help='Path to the metrics definition file')
    parser.add_argument('--prometheus',
                        help='A Prometheus to connect to')
    parser.add_argument('--prometheus-scrape-interval',
                        type=int,
                        help='Prometheus scrape interval (in seconds)',
                        default=15)
    opts, remaining = parser.parse_known_args()
    remaining.insert(0, sys.argv[0])
    TestPrometheus.exporter_path = opts.exporter
    TestPrometheus.exporter_config = opts.config
    TestPrometheus.prometheus = opts.prometheus
    TestPrometheus.prometheus_scrape_interval = opts.prometheus_scrape_interval
    unittest.main(argv=remaining)
